import py2bit
from betterbasequals.utils import p2phred, eprint, reverse_complement, Read, zip_pileups_single_chrom, open_bam_w_index, read_variant_set
from betterbasequals.pilup_handlers import get_pileup_count, get_alleles_w_quals, get_validation_probabities

from collections import Counter

class MutationValidator:
    def __init__(
        self,
        bam_file,
        filter_bam_file,
        validation_bam_file,
        twobit_file,
        kmer_papa
    ):

        # Open files for reading
        self.bam_file = open_bam_w_index(bam_file)
        if filter_bam_file is None:
            self.filter_bam_file = None
        else:
            self.filter_bam_file = open_bam_w_index(filter_bam_file)
        self.validation_bam_file = open_bam_w_index(validation_bam_file)

        self.tb = py2bit.open(twobit_file)
        bquals = list(kmer_papa.keys())
        self.radius = len(next(iter(kmer_papa[bquals[0]]["C->T"].keys())))//2
        self.prefix = ''

        #assumes that the kmer papa has been turned into phred scaled correction factor
        self.mut_probs = kmer_papa
        mtype_tups = ('C->C', ('C->A', 'C->G', 'C->T')), ('A->A', ('A->C', 'A->G', 'A->T'))
        for BQ in self.mut_probs:
            for stay_type, change_types in mtype_tups:
                self.mut_probs[BQ][stay_type] = {}
                for kmer in self.mut_probs[BQ][change_types[0]]:
                    p = 1.0
                    for change_type in change_types:
                        #TODO: should we use log-sum-exp function for numerical stability?
                        #p -= phred2p(self.mut_probs[BQ][change_type][kmer])
                        alpha,beta = self.mut_probs[BQ][change_type][kmer]
                        p -= alpha/(alpha+beta)
                    self.mut_probs[BQ][stay_type][kmer] = (p, None)

        # Create correction factor dict:
        # self.correction_factor = {}
        # for mtype in kmer_papa:
        #     for kmer in kmer_papa[mtype]:
        #         if self.radius is None:
        #             self.radius == len(kmer)//2
        #         p = kmer_papa[mtype][kmer]
        #         self.correction_factor[mtype][kmer] = -10*log10(p/(1-p))
        
        self.min_filter_depth = 25
        self.max_filter_depth = 55


    def __del__(self):
        self.tb.close()

    def call_all_chroms(self):
        for idx_stats in self.bam_file.get_index_statistics():
            if idx_stats.mapped > 0:
                self.call_mutations(idx_stats.contig, None, None)


    def call_mutations(self, chrom, start, stop, mapq=50, mapq_filter=20, min_base_qual_filter=20):
        pileup = self.bam_file.pileup(
            contig = chrom,
            start = start,
            stop = stop,
            truncate = True,
            max_depth = 1000000,
            min_mapping_quality = mapq,
            ignore_overlaps = False,
            flag_require = 0,  # No requirements
            flag_filter = 3840,
            min_base_quality = 2,
        )
        Hifi_pileup = self.validation_bam_file.pileup(
            contig=chrom,
            start=start,
            stop=stop,
            truncate=True,
            min_mapping_quality=mapq_filter,
            flag_filter=3848,
        )
        n_calls = 0
        if not self.filter_bam_file is None:
            filter_pileup = self.filter_bam_file.pileup(
                contig = chrom,
                start = start,
                stop = stop,
                truncate = True,
                min_mapping_quality = mapq_filter,
                ignore_overlaps = False,
                flag_require = 2,  # proper paired
                flag_filter = 3840,
                min_base_quality = min_base_qual_filter,
            )
            for pileupcolumn, hifi_pc, filter_pc in zip_pileups_single_chrom(pileup, Hifi_pileup, filter_pileup):
                n_alleles = Counter()
                for pread in filter_pc.pileups:
                    pos = pread.query_position
                    if not pos is None:
                        n_alleles[pread.alignment.query_sequence[pos]] += 1
                    N_filter = sum(n_alleles.values())
                if N_filter < self.min_filter_depth or N_filter > self.max_filter_depth:
                    continue
                filter_alleles = [x for x,y in n_alleles.items() if y >= 5]
                self.handle_pileup(pileupcolumn, hifi_pc, filter_alleles)

        else:
            for pileupcolumn, hifi_pc in zip_pileups_single_chrom(pileup, Hifi_pileup):
                self.handle_pileup(pileupcolumn, Hifi_pileup, [])


    def handle_pileup(self, pileupcolumn, hifi_pc, filter_alleles):
        ref_pos = pileupcolumn.reference_pos
        chrom = pileupcolumn.reference_name
        if ref_pos%100000 == 0:
            eprint(f"{chrom}:{ref_pos}")            
        #if not self.bed_query_func(chrom, ref_pos):
        #    continue

        if ref_pos-self.radius < 0:
            return 

        kmer = self.tb.sequence(self.prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
        if 'N' in kmer:
            return 
        ref = kmer[self.radius]
        if ref not in "ATGC":
            return 

        if ref not in filter_alleles:
            return 

        hifi_basequals = get_alleles_w_quals(hifi_pc)
        n_hifi_reads = sum(len(hifi_basequals[x]) for x in hifi_basequals)
            
        if n_hifi_reads > 100:
            return

        base_probs, seen_alts, n_mismatch, n_double, n_alt = get_validation_probabities(pileupcolumn, ref, kmer, self.mut_probs)
        
        all_mismatch = sum(n_mismatch.values())
        for A in seen_alts:
            seen_hifi = sum(1 for x in hifi_basequals[A] if x>80) > 0
            for alpha, beta, BQ, muttype, atype in base_probs[A]:
                if n_alt[A]>5:
                    continue
                print(BQ, alpha, beta, muttype, atype, int(seen_hifi), n_mismatch[A], all_mismatch, n_double[A], n_alt[A])

    #             # Variant quality
    #             #corr_var_qual = sum(x[0] for x in corrected_base_quals[A])
    #             #corr_var_qual2 = sum(x[1] for x in corrected_base_quals[A])
    #             #uncorr_var_qual = sum(x[2] for x in corrected_base_quals[A])
    #             n_alt = sum(1 for x in corrected_base_quals[A] if x[1]>35)
    #             seen_hifi = sum(1 for x in hifi_basequals[A] if x>80) > 0
    #             for corrected_Q, uncorrected_Q, base_type in corrected_base_quals[A]:
    #                 print(int(corrected_Q), uncorrected_Q, base_type, seen_hifi, n_mismatch[A], all_mismatch, n_double, n_alt)
    

    # def call_mutations(self, chrom, start, stop):
    #     if self.filter_bam_file is None:
    #         self.call_mutations_no_filter(chrom, start, stop)
    #     else:
    #         self.call_mutations_with_filter(chrom, start, stop)

    # def call_mutations_with_filter(self, chrom, start, stop, mapq=50, mapq_filter=20, min_base_qual_filter=20, prefix=""):
    #     pileup = self.bam_file.pileup(
    #         contig=chrom,
    #         start=start,
    #         stop=stop,
    #         truncate=True,
    #         max_depth = 1000000,
    #         min_mapping_quality=mapq,
    #         ignore_overlaps=False,
    #         flag_require=0,  # proper paired
    #         flag_filter=3848,
    #         min_base_quality = 1,
    #     )
    #     filter_pileup = self.filter_bam_file.pileup(
    #         contig=chrom,
    #         start=start,
    #         stop=stop,
    #         truncate=True,
    #         min_mapping_quality=mapq_filter,
    #         ignore_overlaps=False,
    #         flag_require=2,  # proper paired
    #         flag_filter=3848,
    #     )
    #     Hifi_pileup = self.validation_bam_file.pileup(
    #         contig=chrom,
    #         start=start,
    #         stop=stop,
    #         truncate=True,
    #         min_mapping_quality=mapq_filter,
    #         flag_filter=3848,
    #     )
        
    #     for pileupcolumn, filter_pc, hifi_pc in zip_pileups_single_chrom(pileup, filter_pileup, Hifi_pileup):
    #         ref_pos = pileupcolumn.reference_pos
    #         chrom = pileupcolumn.reference_name            
    #         #if not self.bed_query_func(chrom, ref_pos):
    #         #    continue

    #         kmer = self.tb.sequence(prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
    #         ref = kmer[self.radius]
    #         if 'N' in kmer or len(kmer)!= 2*self.radius +1:
    #             continue            

    #         if ref in ['T', 'G']:
    #             kmer = reverse_complement(kmer)
    #             papa_ref = reverse_complement(ref)
    #         else:
    #             papa_ref = ref

    #         assert len(ref) == 1
    #         if ref not in "ATGC":
    #             continue

    #         n_ref_filter, n_alt_filter = get_pileup_count(filter_pc, ref, min_base_qual_filter, blood=True)
    #         N_filter = n_ref_filter + sum(n_alt_filter.values())
    #         #TODO: replace hardcoded numbers with values relative to mean coverage
    #         if N_filter < 25 or N_filter > 55 or n_ref_filter < 5:
    #             continue

    #         corrected_base_quals, n_ref, n_mismatch, n_double = get_alleles_w_corrected_quals(pileupcolumn, ref, papa_ref, kmer, self.correction_factor)

    #         #n_alt = sum(len(corrected_base_quals[x]) for x in corrected_base_quals)

    #         hifi_basequals = get_alleles_w_quals(hifi_pc)
    #         n_hifi_reads = sum(len(hifi_basequals[x]) for x in hifi_basequals)
            
    #         if n_hifi_reads > 100:
    #             continue

    #         all_mismatch = sum(n_mismatch.values())            

    #         for A in [x for x in ['A','C','G','T'] if x != ref]:    
    #             # Variant quality
    #             #corr_var_qual = sum(x[0] for x in corrected_base_quals[A])
    #             #corr_var_qual2 = sum(x[1] for x in corrected_base_quals[A])
    #             #uncorr_var_qual = sum(x[2] for x in corrected_base_quals[A])
    #             n_alt = sum(1 for x in corrected_base_quals[A] if x[1]>35)
    #             seen_hifi = sum(1 for x in hifi_basequals[A] if x>80) > 0
    #             for corrected_Q, uncorrected_Q, base_type in corrected_base_quals[A]:
    #                 print(int(corrected_Q), uncorrected_Q, base_type, seen_hifi, n_mismatch[A], all_mismatch, n_double, n_alt)
                    
    # def call_mutations_no_filter(self, chrom, start, stop, mapq=50, mapq_hifi=40, prefix=""):
    #     pileup = self.bam_file.pileup(
    #         contig=chrom,
    #         start=start,
    #         stop=stop,
    #         truncate=True,
    #         max_depth = 1000000,
    #         min_mapping_quality=mapq,
    #         ignore_overlaps=False,
    #         flag_require=2,  # proper paired
    #         flag_filter=3848,
    #         min_base_quality = 1,
    #     )
    #     Hifi_pileup = self.validation_bam_file.pileup(
    #         contig=chrom,
    #         start=start,
    #         stop=stop,
    #         truncate=True,
    #         min_mapping_quality=mapq_hifi,
    #         flag_filter=3848,
    #     )
        
    #     for pileupcolumn, hifi_pc in zip_pileups_single_chrom(pileup, Hifi_pileup):
    #         ref_pos = pileupcolumn.reference_pos
    #         chrom = pileupcolumn.reference_name            
    #         #if not self.bed_query_func(chrom, ref_pos):
    #         #    continue

    #         kmer = self.tb.sequence(prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
    #         ref = kmer[self.radius]
    #         if 'N' in kmer or len(kmer)!= 2*self.radius +1:
    #             continue            

    #         if ref in ['T', 'G']:
    #             kmer = reverse_complement(kmer)
    #             papa_ref = reverse_complement(ref)
    #         else:
    #             papa_ref = ref

    #         assert len(ref) == 1
    #         if ref not in "ATGC":
    #             continue

    #         corrected_base_quals, n_ref, n_mismatch, n_double = get_alleles_w_corrected_quals(pileupcolumn, ref, papa_ref, kmer, self.correction_factor)

    #         #n_alt = sum(len(corrected_base_quals[x]) for x in corrected_base_quals)

    #         hifi_basequals = get_alleles_w_quals(hifi_pc)
    #         #n_hifi_reads = sum(len(hifi_basequals[x]) for x in hifi_basequals)
    #         all_mismatch = sum(n_mismatch.values())            

    #         for A in [x for x in ['A','C','G','T'] if x != ref]:    
    #             # Variant quality
    #             #corr_var_qual = sum(x[0] for x in corrected_base_quals[A])
    #             #corr_var_qual2 = sum(x[1] for x in corrected_base_quals[A])
    #             #uncorr_var_qual = sum(x[2] for x in corrected_base_quals[A])

    #             for corrected_Q, uncorrected_Q, base_type in corrected_base_quals[A]:
    #                 print(int(corrected_Q), uncorrected_Q, base_type, sum(1 for x in hifi_basequals[A] if x>80), n_mismatch[A], all_mismatch, n_double)
    #                 #print(chrom, ref_pos, A, int(corrected_Q), uncorrected_Q, base_type, sum(hifi_basequals[A]), n_mismatch, n_hifi_reads)



class ListMutationValidator:
    def __init__(
        self,
        bam_file,
        filter_bam_file,
        validation_list_file, 
        twobit_file,
        kmer_papa
    ):

        # Open files for reading
        self.bam_file = open_bam_w_index(bam_file)
        if filter_bam_file is None:
            self.filter_bam_file = None
        else:
            self.filter_bam_file = open_bam_w_index(filter_bam_file)
        #self.validation_bam_file = open_bam_w_index(validation_bam_file)

        self.validation_set = read_variant_set(validation_list_file) 

        self.tb = py2bit.open(twobit_file)
        bquals = list(kmer_papa.keys())
        self.radius = len(next(iter(kmer_papa[bquals[0]]["C->T"].keys())))//2
        self.prefix = ''

        #assumes that the kmer papa has been turned into phred scaled correction factor
        self.mut_probs = kmer_papa
        # mtype_tups = ('C->C', ('C->A', 'C->G', 'C->T')), ('A->A', ('A->C', 'A->G', 'A->T'))
        # for BQ in self.mut_probs:
        #     for stay_type, change_types in mtype_tups:
        #         self.mut_probs[BQ][stay_type] = {}
        #         for kmer in self.mut_probs[BQ][change_types[0]]:
        #             p = 1.0
        #             for change_type in change_types:
        #                 #TODO: should we use log-sum-exp function for numerical stability?
        #                 #p -= phred2p(self.mut_probs[BQ][change_type][kmer])
        #                 alpha,beta = self.mut_probs[BQ][change_type][kmer]
        #                 p -= alpha/(alpha+beta)
        #             self.mut_probs[BQ][stay_type][kmer] = (p, None)


        # Create correction factor dict:
        # self.correction_factor = {}
        # for mtype in kmer_papa:
        #     for kmer in kmer_papa[mtype]:
        #         if self.radius is None:
        #             self.radius == len(kmer)//2
        #         p = kmer_papa[mtype][kmer]
        #         self.correction_factor[mtype][kmer] = -10*log10(p/(1-p))
        
        self.min_filter_depth = 25
        self.max_filter_depth = 55


    def __del__(self):
        self.tb.close()

    def call_all_chroms(self):
        for idx_stats in self.bam_file.get_index_statistics():
            if idx_stats.mapped > 0:
                self.call_mutations(idx_stats.contig, None, None)


    def call_mutations(self, chrom, start, stop, mapq=50, mapq_filter=20, min_base_qual_filter=20):
        pileup = self.bam_file.pileup(
            contig = chrom,
            start = start,
            stop = stop,
            truncate = True,
            max_depth = 1000000,
            min_mapping_quality = mapq,
            ignore_overlaps = False,
            flag_require = 0,  # No requirements
            flag_filter = 3840,
            min_base_quality = 2,
        )
        n_calls = 0
        if not self.filter_bam_file is None:
            filter_pileup = self.filter_bam_file.pileup(
                contig = chrom,
                start = start,
                stop = stop,
                truncate = True,
                min_mapping_quality = mapq_filter,
                ignore_overlaps = False,
                flag_require = 2,  # proper paired
                flag_filter = 3840,
                min_base_quality = min_base_qual_filter,
            )
            for pileupcolumn, filter_pc in zip_pileups_single_chrom(pileup, filter_pileup):
                n_alleles = Counter()
                for pread in filter_pc.pileups:
                    pos = pread.query_position
                    if not pos is None:
                        n_alleles[pread.alignment.query_sequence[pos]] += 1
                    N_filter = sum(n_alleles.values())
                if N_filter < self.min_filter_depth or N_filter > self.max_filter_depth:
                    continue
                filter_alleles = [x for x,y in n_alleles.items() if y >= 5]
                self.handle_pileup(pileupcolumn, filter_alleles)

        else:
            for pileupcolumn in pileup:
                self.handle_pileup(pileupcolumn, None)


    def handle_pileup(self, pileupcolumn, filter_alleles):
        ref_pos = pileupcolumn.reference_pos
        chrom = pileupcolumn.reference_name
        if ref_pos%100000 == 0:
            eprint(f"{chrom}:{ref_pos}")            
        #if not self.bed_query_func(chrom, ref_pos):
        #    continue

        if ref_pos-self.radius < 0:
            return 

        kmer = self.tb.sequence(self.prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
        if 'N' in kmer:
            return
        if len(kmer) != 2*self.radius + 1:
            return
        ref = kmer[self.radius]
        if ref not in "ATGC":
            return 

        if not filter_alleles is None and ref not in filter_alleles:
            return 

        ## FIXME: get_alleles_w_probabities_update_ver2 is outdated and should be updated before it is used:
        #base_probs, n_mismatch, n_double, n_mismatch_BQ, n_double_BQ = get_alleles_w_probabities_update_ver2(pileupcolumn, ref, kmer, self.mut_probs, 100, True, 0.5)
        #for A in base_probs:
        #    seen_validation = (chrom, ref_pos+1, A) in self.validation_set
        #    N_A = sum(oldBQ > 30 for newBQ, MQ, oldBQ in base_probs[A])#sum(1 for x,y,_ in base_probs[A] if x<y)
        #    for newBQ, MQ, oldBQ in base_probs[A]:            
        #        print(oldBQ, int(newBQ), int(seen_validation), n_mismatch[A], n_double[A], n_mismatch_BQ[A][oldBQ], n_double_BQ[A][oldBQ], N_A, int(MQ))


        # base_probs, seen_alts, n_mismatch, n_double, n_alt = get_validation_probabities(pileupcolumn, ref, kmer, self.mut_probs)    
        # #print(len(base_probs), len(seen_alts))
        # all_mismatch = sum(n_mismatch.values())
        # for A in seen_alts:
        #     seen_validation = (chrom, ref_pos+1, A) in self.validation_set
        #     for alpha, beta, BQ, muttype, atype in base_probs[A]:
        #         #if n_alt[A]>5:
        #         #    continue
        #         print(BQ, alpha, beta, muttype, atype, int(seen_validation), n_mismatch[A], all_mismatch, n_double[A], n_alt[A])



import pysam
import py2bit
import os
from collections import Counter
from betterbasequals.utils import reverse_complement, Read, zip_pileups_single_chrom, eprint, open_bam_w_index, SortedBed

def get_mut_type(ref, alt):
    if ref in ['T', 'G']:
        mtype = reverse_complement(ref) + '->' + reverse_complement(alt)
    else:
        mtype = ref + '->' + alt
    return mtype

def get_BQ_pair(BQ1, BQ2):
    if BQ1 > BQ2:
        return f'({BQ1},{BQ2})'
    else:
        return f'({BQ2},{BQ1})'

class MutationCounterWFilter:
    def __init__(
        self,
        bam_file,
        twobit_file,
        filter_bam_file,
        mapq = 50,
        min_base_qual = 1,
        filter_mapq = 20,
        min_base_qual_filter=20, 
        min_depth=1, 
        max_depth=1000000, 
        radius=3, 
        prefix="", 
        max_alt_frac = None,
        min_filter_count = 2,
        min_filter_depth = 10,
        max_filter_depth = 100,
        min_enddist = 6,
        max_mismatch = 2,
        bed_file = None,
        #output,
    ):
        self.bam_file = open_bam_w_index(bam_file)
        if filter_bam_file is None:
            self.filter_bam_file = None
            if max_alt_frac is None:
                #Set default to filter heterozygous variants if no filter is used:
                self.max_alt_frac = 0.2
        else:
            self.filter_bam_file = open_bam_w_index(filter_bam_file)
            if max_alt_frac is None:
                #Should I have different default here?
                #Maybe 1.0 corresponding to no filter?
                self.max_alt_frac = 0.2

        self.tb = py2bit.open(twobit_file)
   
        self.mapq = mapq
        self.min_base_qual = min_base_qual
        self.filter_mapq = filter_mapq
        self.min_base_qual_filter = min_base_qual_filter
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.radius = radius
        self.prefix = prefix
        self.min_filter_count = min_filter_count
        self.min_filter_depth = min_filter_depth
        self.max_filter_depth = max_filter_depth
        self.min_enddist = min_enddist
        self.max_mismatch = max_mismatch
        if not bed_file is None:
            self.bed = SortedBed(bed_file)
        else:
            self.bed = None

    def __del__(self):
        self.tb.close()

    def count_mutations_all_chroms(self):
        event_kmers = Counter()
        for idx_stats in self.bam_file.get_index_statistics():
            if idx_stats.mapped > 0:
                event_c = self.count_mutations(idx_stats.contig, None, None)
                event_kmers += event_c
        return event_kmers


    def count_mutations(self, chrom, start, stop):
        event_kmers = Counter()

        pileup = self.bam_file.pileup(
            contig = chrom,
            start = start,
            stop = stop,
            truncate = True,
            max_depth = 1000000,
            min_mapping_quality = self.mapq,
            ignore_overlaps = False,
            flag_require = 0,  # No requirement
            flag_filter = 3848,
            min_base_quality = self.min_base_qual,
        )
        if not self.filter_bam_file is None:
            filter_pileup = self.filter_bam_file.pileup(
                contig = chrom,
                start = start,
                stop = stop,
                truncate = True,
                min_mapping_quality = self.filter_mapq,
                ignore_overlaps = False,
                flag_require = 2,  # proper paired
                flag_filter = 3848,
                min_base_quality = self.min_base_qual_filter,
            )
            for pileupcolumn, filter_pc in zip_pileups_single_chrom(pileup, filter_pileup):
                N_filter = 0
                n_alleles = Counter()
                for pread in filter_pc.pileups:
                    pos = pread.query_position
                    if not pos is None:
                        n_alleles[pread.alignment.query_sequence[pos]] += 1
                        N_filter += 1
                if N_filter < self.min_filter_depth or N_filter > self.max_filter_depth:
                    continue                

                # We can set filter so that we only considder sites where
                # Any non-ref allele is seen min_filter_count times or more
                if len(n_alleles) > 1:
                    major, second = n_alleles.most_common(2)
                    if second[1] >= self.min_filter_count:
                        continue
                else:
                    major = n_alleles.most_common(1)[0]

                self.handle_pileup(pileupcolumn, event_kmers, major[0])
        else:
            for pileupcolumn in pileup:
                self.handle_pileup(pileupcolumn, event_kmers)
        
        return event_kmers

    def handle_pileup(self, pileupcolumn, event_kmers, major_allele = None):
        ref_pos = pileupcolumn.reference_pos
        chrom = pileupcolumn.reference_name
        #if ref_pos%10000 ==0:
        #    eprint(f"{chrom}:{ref_pos}")    

        if not self.bed is None and self.bed.query(chrom, ref_pos):
            return

        if ref_pos - self.radius < 0:
            return 

        kmer = self.tb.sequence(self.prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
        
        if 'N' in kmer:
            return
        
        if len(kmer) != 2*self.radius + 1:
            return 0
        
        ref = kmer[self.radius]

        #If the major in the filter bam file does not match the ref we ignore the site.
        if major_allele is not None and major_allele != ref:
            return

        if ref not in "ATGC":
            return
        
        if ref not in ['A', 'C']:
            kmer = reverse_complement(kmer)

        reads_mem = {}
        event_list = []
        coverage = 0

        has_good = set()
        n_alleles = Counter()

        for pileup_read in pileupcolumn.pileups:
            # test for deletion at pileup
            #if pileup_read.is_del or pileup_read.is_refskip:
            #    continue
            coverage += 1
            #TODO: should consider what the right solution is if there is deletion at overlap

            # fetch read information
            read = Read(pileup_read)
            
            # # test if read is okay
            # if (
            #     read.allel not in "ATGC"
            #     or read.start is None
            #     or read.end is None
            #     #or read.NH != 1
            # ):
            #     continue

            if read.base_qual > 30:
                n_alleles[read.allel] += 1

            # Look for read partner
            if read.query_name in reads_mem:
                # found partner process read pair
                mem_read = reads_mem.pop(read.query_name)
                
                if (not read.is_good(self.min_enddist, self.max_mismatch, self.mapq)) or (not mem_read.is_good(self.min_enddist, self.max_mismatch, self.mapq)):
                    continue

                if read.allel == mem_read.allel:
                    event_list.append(('good', read.allel, read.base_qual))
                    event_list.append(('good', mem_read.allel, mem_read.base_qual))
                    event_list.append(('good_tuple', read.allel, get_BQ_pair(read.base_qual, mem_read.base_qual)))

                    if max(read.base_qual, mem_read.base_qual) > 30:
                        has_good.add(read.allel)

                else:
                    if read.allel != ref and mem_read.allel == ref:# and mem_read.base_qual > 30:
                        event_list.append(('bad', read.allel, read.base_qual))
                        event_list.append(('good', mem_read.allel, mem_read.base_qual))
                        event_list.append(('bad_tuple', read.allel, get_BQ_pair(read.base_qual, mem_read.base_qual)))

                    elif mem_read.allel != ref and read.allel == ref:# and read.base_qual > 30:
                        event_list.append(('bad', mem_read.allel, mem_read.base_qual))
                        event_list.append(('good', read.allel, read.base_qual))
                        event_list.append(('bad_tuple', read.allel, get_BQ_pair(read.base_qual, mem_read.base_qual)))

            else:            
                reads_mem[read.query_name] = read
        
        # Handle reads without partner (ie. no overlap)
        for read in reads_mem.values():
            if read.is_good(self.min_enddist, self.max_mismatch, self.mapq):
                event_list.append(('singleton', read.allel, read.base_qual))

        if coverage < self.min_depth or coverage > self.max_depth:
            return
        
        N = sum(n_alleles.values())

        # We ignore sites where matching reference allele aren't seen in an overlap
        # OBS: Have commented out. 2023.06.02
        # In files with litlle overap we maybe should not require this.
        #if ref not in has_good:
        #    return
                
        if len(n_alleles) > 1:
            major, second = n_alleles.most_common(2)
            # If we see a high fraction of an alternative allele we ignore the site.
            if second[1]/N >= self.max_alt_frac:
                return
        elif len(n_alleles) == 1:
            major = n_alleles.most_common(1)[0]
        else:
            #We require that we see a good quality allele at the site
            return

        #If the major allele is not the ref allele we ignore the site.
        if major[0] != ref:
            return

        for event_type, allele, base_qual in event_list:
            mut_type = get_mut_type(ref, allele)     
            
            event_kmers[(event_type, base_qual, mut_type, kmer)] += 1




from math import sqrt


def get_error_probs(a, b, c):
    """calculate error probabilities for the three different
    mutation types at a type of site (A or C)

    Args:
        a (Float): Rate of incompatibilities of first type fx. n(A->A/C, incomp)/n(A->A/?)
        b (Float): Rate of incompatibilities of second type fx. n(A->A/G, incomp)/n(A->A/?)
        c (Float): Rate of incompatibilities of third type fx. n(A->A/T, incomp)/n(A->A/?)
    """
    # No error prob:
    abc_sum = a + b + c
    d = 1 - 2*abc_sum
    p_AA = 0.5 + sqrt(d)/2
    p_AC = a/(2*p_AA)
    p_AG = b/(2*p_AA)
    p_AT = c/(2*p_AA)
    print(d)
    print(p_AA)
    return p_AC, p_AG, p_AT
sum([x for x in get_error_probs(1/100,1/1000,1/1000)])


def get_error_probs(a, b, c):
    """calculate error probabilities for the three different
    mutation types at a type of site (A or C)

    Args:
        a (Float): Rate of incompatibilities of first type fx. n(A->C, incomp)/n(A->A match)
        b (Float): Rate of incompatibilities of second type fx. n(A->G, incomp)/n(A->A match)
        c (Float): Rate of incompatibilities of third type fx. n(A->T, incomp)/n(A->A match)
    """

    # No error prob:
    abc_sum = a + b + c
    d = 1 - 2*abc_sum
    #print(f'abc_sum={abc_sum} d={d}')
    p_no_error_1 = 1- ((-1 + sqrt(d)) / -2)
    p_no_error_2 = 1- ((-1 - sqrt(d)) / -2)
    #print(p_no_error_1)
    #print(p_no_error_2)
    p_no_error= max(p_no_error_1, p_no_error_2)
    pa = a/(2*p_no_error)
    pb = b/(2*p_no_error)
    pc = c/(2*p_no_error)
    print(pa,pb,pc)
    #print(b/p_no_error)    
    #print(c/p_no_error)
    



"""Module that contains the command line application."""

import argparse
from betterbasequals.get_good_bad_kmers import MutationCounterWFilter
from betterbasequals.model_validators import MutationValidator, ListMutationValidator
from betterbasequals.bam_adjusters import BaseAdjuster
from betterbasequals.somatic_callers import SomaticMutationCaller
from betterbasequals.utils import *
from betterbasequals import __version__
from betterbasequals.kmerpapa_utils import get_kmerpapa
from betterbasequals.filter_calls import BBQFilter
from math import sqrt

def get_parser():
    """
    Return the CLI argument parser.

    Returns:
        An argparse parser.
    """
    parser = argparse.ArgumentParser(
        prog="bbq",
        description='''
        Calculates sample-specific base qualities using overlapping reads.
        ''',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )

    # top level args:
    parser.add_argument("--verbosity", type=int, default=1)
    parser.add_argument('-V', '--version', action='version', version=f'%(prog)s {__version__}')

    subparsers = parser.add_subparsers(dest='command',
        title="commands",
        help='Select one of:')

    # args bam file:
    bam_parent = argparse.ArgumentParser(add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    bam_parent.add_argument("--bam_file", required=True,
        help="bam file")
    bam_parent.add_argument("--twobit_file", required=True,
        help="Reference genome in two-bit format")
    bam_parent.add_argument('--region', '-r', type=str,
        help='only consider variants in this region')

    # args for filter bam file
    filter_parent = argparse.ArgumentParser(add_help=False)
    filter_parent.add_argument("--filter_bam_file", help="filter bam file, variants present in this file will not be called")
    
    read_filter_parent = argparse.ArgumentParser(add_help=False, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    read_filter_parent.add_argument('--min_enddist', type=int, default=5, metavar="M",
        help="Ignore bases in the first M or last M positions in the read")
    read_filter_parent.add_argument('--max_mismatch', type=int, default=2, metavar="M",
        help="Ignore alt reads if the read has more than M mismatches to the reference")
    read_filter_parent.add_argument('--min_MQ', type=int, default=50,
        help="Minimum base quality to considder")
    read_filter_parent.add_argument('--exclude_bed', type=str,
        help='Bed file with regions to exclude. Should be sorted by chrom (alphabetically) then pos (numrically).')

    # args for counting kmers:
    count_parent = argparse.ArgumentParser(add_help=False)
    count_parent.add_argument("--output_file_kmers", type=argparse.FileType('w'),
        help="file to save the kmer counts in")
    count_parent.add_argument("--radius", type=int, default=3, 
        help="number of bases around a position that should be considered")
    count_parent.add_argument('--min_depth', type=int, default=1,
        help="mminimum depth at a site to be considered as training data")
    count_parent.add_argument('--max_depth', type=int, default=5000,
        help="maximum depth at a site to be considered as training data")
    count_parent.add_argument('--filter_min_depth', type=int, default=1,
        help="minimum depth in filter_bam_file for a site to be considered as training data")
    count_parent.add_argument('--filter_max_depth', type=int, default=5000,
        help="maximum depth om filter_bam_file at a site to be considered as training data")


    # args for training models:    
    train_parent = argparse.ArgumentParser(add_help=False)
    train_parent.add_argument('--kmerpapa_method', type=str, default = "optimal",
        help='algorithm to use for calculating kmer-papas',
        choices=['greedy', 'optimal'])
    train_parent.add_argument('--correction_type', type=str, default = "bad_vs_no",
        help='should we compare bad variants to "good variants"(SNVs) or to "no variant" (homozygous ref sites)',
        choices=["bad_vs_good", "bad_vs_no"])
    train_parent.add_argument("--output_file_kmerpapa", type=argparse.FileType('w'),
        help="save kmerpapa model to file")
    train_parent.add_argument("--output_file_EQ", type=argparse.FileType('w'))
    train_parent.add_argument("--CVfile_prefix", type=str)
    train_parent.add_argument('-N', '--nfolds', type=int, metavar='N', default=2,
        help='Number of folds to use when fitting hyperparameters in kmerpapa')
    train_parent.add_argument('-i', '--iterations', type=int, default=1, metavar='i',
        help='Repeat cross validation i times when fitting hyperparameters in kmerpapa')
    train_parent.add_argument('--seed', type=int,
        help='seed for numpy.random')
    train_parent.add_argument('--same_good', action='store_true')
    train_parent.add_argument('--subtract', action='store_true')
    train_parent.add_argument('--no_EQ_pat', action='store_true')
    train_parent.add_argument('--EQ_pat_alpha', type=float, default=2)
    train_parent.add_argument('--min1EQ', action='store_true')
    train_parent.add_argument('--estimated', type=str, default = 'single',
                              choices = ['single', 'double'])
    train_parent.add_argument('--mean_type', type=str, default = 'geometric',
                              choices = ['geometric', 'harmonic', 'arithmetic'])
    train_parent.add_argument(
        '-a', '--pseudo_counts', type=float, metavar='a', nargs='+', default = [1,10,30],
        help='Different pseudo count (alpha) values to test using cross validation')
    train_parent.add_argument(
        '-c', '--penalty_values', type=float, metavar='c', nargs='+', default = [5,10,15,20,25,30,35],
        help='Different penalty values to test using cross validation.')

    # args for validating models:    
    validate_parent = argparse.ArgumentParser(add_help=False)
    validate_parent.add_argument("--validation_bam_file", help="hifi bam file", required=True)

    # args for validating models:    
    list_validate_parent = argparse.ArgumentParser(add_help=False)
    list_validate_parent.add_argument("--validation_list_file", help="file with list of true variants", required=True)

    # args for printing polished bam:    
    adjust_parent = argparse.ArgumentParser(add_help=False)
    adjust_parent.add_argument('--outbam', type=str,
        help="Bam file with adjusted base qualities.", required=True)
    adjust_parent.add_argument("--output_adjustments", type=argparse.FileType('w'))
    adjust_parent.add_argument("--cutoff", type=int,
        help="set adjusted basequalities lower than this cutoff to zero")

    # args for calling somatic variants:    
    call_parent = argparse.ArgumentParser(add_help=False)
    call_parent.add_argument('--outfile', type=argparse.FileType('w'), default=sys.stdout,
        help="output file")
    call_parent.add_argument('--method', type=str,
        choices=['LR', 'LR_with_MQ', 'maxLR_with_MQ', 'BF', 'BF_with_MQ', 'BF_with_MQ_and_Prior'], default="LR_with_MQ",
        help="Method used to calculate variant quality scores")
    call_parent.add_argument('--cutoff', type=float, default=None, metavar='Q',
        help="Only print variants with quality above Q.")
    call_parent.add_argument('--prior_N', type=float, default=50000,
        help="Weight (as sample size) of the kmer based prior on error rate.")
    call_parent.add_argument('--no_update',  action='store_true',
        help="Do not make bayesian update of error rate but use rate only estimated from kmers")
    call_parent.add_argument('--min_BQ', type=int, default=1,
        help="Minimum base quality to considder")
    call_parent.add_argument('--filter_max_count', type=int, default=2,
        help='Maximum number of times an alternative read is allowed to be seen in filer_bam')
    call_parent.add_argument("--pop_vcf", type=str,
        help='Population vcf with AF field.')
    call_parent.add_argument('--max_NM_diff', type=int, default=2,
        help="Maximum allowed difference in median number of mismatches for reads with alt alleles compared to reads with ref alleles")

    count_parser = subparsers.add_parser('count', 
        description='Count good and bad k-mers',
        help =  'Count good and bad k-mers',
        parents=[bam_parent, filter_parent, count_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    train_parser = subparsers.add_parser('train', 
        description='First run "count" then train model to distinguish good and bad k-mers.', 
        help = 'First run "count" then train model to distinguish good and bad k-mers.',
        parents=[bam_parent, filter_parent, count_parent, train_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    validate_parser = subparsers.add_parser('validate', 
        description = 'First run "count" and "train" then print validation data',
        help = 'First run "count" and "train" then print validation data',
        parents=[bam_parent, filter_parent, count_parent, train_parent, validate_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    list_validate_parser = subparsers.add_parser('list_validate', 
        description = 'First run "count" and "train" then print validation data',
        help = 'First run "count" and "train" then print validation data',
        parents=[bam_parent, filter_parent, count_parent, train_parent, list_validate_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # adjust_parser = subparsers.add_parser('adjust', 
    #     description = 'First run "count" and "train" then output bam with adjusted base qualities.', 
    #     help = 'First run "count" and "train" then output bam with adjusted base qualities.', 
    #     parents = [bam_parent, filter_parent, count_parent, train_parent, adjust_parent],
    #     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    call_parser = subparsers.add_parser('call', 
        description = 'First run "count" and "train" then call variants', 
        help = 'First run "count" and "train" then call variants', 
        parents = [bam_parent, filter_parent, count_parent, train_parent, call_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    train_only_parser = subparsers.add_parser('train_only', 
        description = 'Train model to distinguish good and bad k-mers.',
        help = 'Train model to distinguish good and bad k-mers.',
        parents = [train_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    train_only_parser.add_argument("--input_file_kmers", type=argparse.FileType('r'))

    validate_only_parser = subparsers.add_parser('validate_only', 
        description = 'Print validation data.',
        help = 'Print validation data.', 
        parents = [bam_parent, filter_parent, validate_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter) 
    validate_only_parser.add_argument("--input_file_kmerpapa", type=argparse.FileType('r'))

    list_validate_only_parser = subparsers.add_parser('list_validate_only', 
        description = 'Print validation data.',
        help = 'Print validation data.', 
        parents = [bam_parent, filter_parent, list_validate_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter) 
    list_validate_only_parser.add_argument("--input_file_kmerpapa", type=argparse.FileType('r'))

    # adjust_only_parser = subparsers.add_parser('adjust_only', 
    #     description = 'Output bam with adjusted base qualities.', 
    #     help = 'Output bam with adjusted base qualities.',
    #     parents = [bam_parent, adjust_parent],
    #     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # adjust_only_parser.add_argument("--input_file_kmerpapa", type=argparse.FileType('r'))

    call_only_parser = subparsers.add_parser('call_only', 
        description = 'Call variants',
        help = 'Call variants',
        parents=[bam_parent, filter_parent, call_parent, read_filter_parent],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    call_only_parser.add_argument("--input_file_kmerpapa", type=argparse.FileType('r'))
    
    # test_kmerpapa_parser = subparsers.add_parser('test_kmerpapa', 
    #     description = 'Apply a kmerpapa model to a set of kmer counts',
    #     help = 'Apply a kmerpapa model to a set of kmer counts',
    #     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # test_kmerpapa_parser.add_argument("--input_file_kmers", type=argparse.FileType('r'))
    # test_kmerpapa_parser.add_argument("--input_file_kmerpapa", type=argparse.FileType('r'))
    # test_kmerpapa_parser.add_argument('--correction_type', type=str, default = "bad_vs_no",
    #     help='should we compare bad variants to "good variants"(SNVs) or to "no variant" (homozygous ref sites)',
    #     choices=["bad_vs_good", "bad_vs_no"])
    # test_kmerpapa_parser.add_argument('--same_good', action='store_true')

    filter_calls_parser = subparsers.add_parser('filter_calls', 
        description = 'Filter PASS calls based on coverage quantiles',
        help = 'Filter PASS calls based on coverage quantiles',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    filter_calls_parser.add_argument('--vcf_file', required=True,
        help='input VCF file')
    filter_calls_parser.add_argument('--outfile',  type=argparse.FileType('w'), default=sys.stdout,
        help='output VCF file') 
    filter_calls_parser.add_argument('--lower_q', type=float, default=0.05,
        help='Lower quantile, PASS calls with coverage below this value will be filtered out') 
    filter_calls_parser.add_argument('--upper_q', type=float, default=0.95,
        help='Upper quantile, PASS calls with coverage above this value will be filtered out') 
    return parser


def run_get_good_and_bad_w_filter(opts):
    if opts.verbosity > 0:
        eprint("Counting good and bad kmers")
    counter = \
        MutationCounterWFilter(
            opts.bam_file, 
            opts.twobit_file, 
            opts.filter_bam_file,
            opts.min_MQ,
            radius = opts.radius,
            min_enddist = opts.min_enddist,
            max_mismatch = opts.max_mismatch,
            bed_file = opts.exclude_bed)
    
    if opts.chrom is None:
        event_kmers = counter.count_mutations_all_chroms()
    else:
        event_kmers = counter.count_mutations(opts.chrom, opts.start, opts.end)
    print_kmer_counts(opts, event_kmers)
    
    return event_kmers
    

def run_get_kmerpapas(opts, event_kmers):
    if opts.verbosity > 0:
        eprint("Training kmer pattern partitions")
    kmer_papas = {}
    kmer_patterns = {}
    BQs = list(set(x[1] for x in event_kmers if x[0]in ['good', 'bad']))
    BQs.sort()
    BQs = BQs[::-1]
    radius = len(next(iter(event_kmers.keys()))[3])//2
    for bqual in BQs:
        BQ_pair = f'({bqual},{bqual})'
        kmer_papas[BQ_pair] = {}
        kmer_papas[bqual] = {}
        kmer_patterns[bqual] = {}
        for mtype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T'):
            contextD = {}
            ref = mtype[0]
            alt = mtype[-1]
            super_pattern = 'N'*radius + ref + 'N'*radius
            notype = f'{ref}->{ref}'
            other_type1, other_type2 = [f'{ref}->{other}' for other in 'ACGT' if other != alt and other != ref]
            eprint(f'Handling base_qual: {BQ_pair}, muttype: {mtype}')
            for kmer in matches(super_pattern):
                n_errors = event_kmers[('bad', bqual, mtype, kmer)]
                n_not_error = event_kmers[('good', bqual, notype, kmer)]
                if not opts.same_good:
                    n_not_error += event_kmers[('bad', bqual, other_type1, kmer)]
                    n_not_error += event_kmers[('bad', bqual, other_type2, kmer)]
                contextD[kmer] = (n_errors, n_not_error)

            kpp = get_kmerpapa(super_pattern, contextD, f'{mtype}_{bqual}', opts)
            if opts.estimated == "single":
                estBQ = bqual
            elif opts.estimated == "double":
                estBQ = BQ_pair

            kmer_papas[bqual][mtype] = {}    
            kmer_papas[BQ_pair][mtype] = {}
            
            kmer_patterns[bqual][mtype] = kpp
            for pat in kpp:
                alpha, beta = kpp[pat]
                p_error = min(0.25, alpha / (alpha + beta))
                if opts.verbosity > 1:
                    eprint(estBQ, mtype, pat, p_error, -10*log10(p_error))
                for kmer in matches(pat):
                    kmer_papas[estBQ][mtype][kmer] = p_error

    # Estimate EQ correction for bases in overlaps
    single_EQ = {}
    double_EQ = {}
    avg_single_EQ = {}
    avg_double_EQ = {}
    for BQ in BQs:
        single_EQ[BQ] = {}
        double_EQ[BQ] = {}
        avg_single_EQ[BQ] = {}
        avg_double_EQ[BQ] = {}
        for mtype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T'):
            ref = mtype[0]
            alt = mtype[-1]
            super_pattern = 'N'*radius + ref + 'N'*radius
            notype = f'{ref}->{ref}'
            other_type1, other_type2 = [f'{ref}->{other}' for other in 'ACGT' if other != alt and other != ref]
            BQ_pair = f'({BQ},{BQ})'
            single_mut = 0
            single_nomut = 0
            double_mut = 0
            double_nomut = 0
            if opts.estimated == 'single':
                BQest = BQ
            elif opts.estimated == 'double':
                BQest = BQ_pair
            
            if not opts.no_EQ_pat:
                total_single_mut = 0
                total_single_nomut = 0
                total_double_mut = 0
                total_double_nomut = 0
                single_EQ[BQ][mtype] = {}
                double_EQ[BQ][mtype] = {}
                avg_single_EQ[BQ][mtype] = {}
                avg_double_EQ[BQ][mtype] = {}
                for pat in kmer_patterns[BQ][mtype]:
                    single_mut = 0
                    single_nomut = 0
                    double_mut = 0
                    double_nomut = 0
                    for kmer in matches(pat):
                        single_mut += event_kmers[('singleton', BQ, mtype, kmer)]
                        single_nomut += event_kmers[('singleton', BQ, notype, kmer)]
                        double_mut += event_kmers[('good_tuple', BQ_pair, mtype, kmer)]
                        double_nomut += event_kmers[('good_tuple', BQ_pair, notype, kmer)]
                        #TODO: add other muttypes from good_tuple and single to nomut?
                        double_nomut += event_kmers[('bad_tuple', BQ_pair, mtype, kmer)]
                        double_nomut += event_kmers[('bad_tuple', BQ_pair, other_type1, kmer)]
                        double_nomut += event_kmers[('bad_tuple', BQ_pair, other_type2, kmer)]
                    total_single_mut += single_mut
                    total_single_nomut += single_nomut
                    total_double_mut += double_mut
                    total_double_nomut += double_nomut
                    single_EQ[BQ][mtype][pat] = (single_mut, single_nomut)#(single_mut+0.1) / (single_mut + single_nomut + 0.2)
                    double_EQ[BQ][mtype][pat] = (double_mut, double_nomut)#(double_mut+0.1) / (double_mut + double_nomut + 0.2)
                avg_single_EQ[BQ][mtype] = (total_single_mut + 0.1) / (total_single_mut + total_single_nomut + 0.2)
                avg_double_EQ[BQ][mtype] = (total_double_mut + 0.1) / (total_double_mut + total_double_nomut + 0.2)

            else:
                for kmer in kmer_papas[BQest][mtype]:
                    single_mut += event_kmers[('singleton', BQ, mtype, kmer)]
                    single_nomut += event_kmers[('singleton', BQ, notype, kmer)]
                    double_mut += event_kmers[('good_tuple', BQ_pair, mtype, kmer)]
                    double_nomut += event_kmers[('good_tuple', BQ_pair, notype, kmer)]
                    #TODO: add other muttypes from good_tuple and single to nomut?
                    double_nomut += event_kmers[('bad_tuple', BQ_pair, mtype, kmer)]
                    double_nomut += event_kmers[('bad_tuple', BQ_pair, other_type1, kmer)]
                    double_nomut += event_kmers[('bad_tuple', BQ_pair, other_type2, kmer)]
                single_EQ[BQ][mtype] = (single_mut+0.1) / (single_mut + single_nomut + 0.2)
                double_EQ[BQ][mtype] = (double_mut+0.1) / (double_mut + double_nomut + 0.2)
           
           # rel_EQ = single_EQ / double_EQ

            #if not opts.output_file_EQ is None:
            #    print(BQ, mtype, single_mut, single_nomut, single_EQ, double_mut, double_nomut, double_EQ, rel_EQ, file = opts.output_file_EQ)

    #if not opts.output_file_EQ is None:
    #    opts.output_file_EQ.close()        

    if not opts.output_file_EQ is None:
        if not opts.no_EQ_pat:
            print('BQ mutationtype pattern kmerpapa_alpha kmepapa_beta kmerpapa_rate single_mut single_nomut double_mut double_nomut single_EQ double_EQ subtract relative_EQ', file = opts.output_file_EQ)



    max_BQ = max(BQs)
    alpha = opts.EQ_pat_alpha #5 # pseudo count
    # apply EQ correction
    for BQ in BQs:
        for mtype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T'):
            BQ_pair = f'({BQ},{BQ})'
            if (opts.subtract and
                (BQ < max_BQ) and 
                (single_EQ[BQ][mtype] > double_EQ[max_BQ][mtype] + 1e-8) and 
                (double_EQ[BQ][mtype] > double_EQ[max_BQ][mtype] + 1e-8)):
                subtract = double_EQ[max_BQ][mtype]
            else:
                subtract = 0.0

            if opts.no_EQ_pat:
                rel_EQ = (single_EQ[BQ][mtype] - subtract) / (double_EQ[BQ][mtype] - subtract)
                if not opts.output_file_EQ is None:
                    print(BQ, mtype, single_EQ[BQ][mtype], double_EQ[BQ][mtype], subtract, rel_EQ, file = opts.output_file_EQ)

            if opts.estimated == "single":
                kmer_papas[BQ_pair][mtype] = {}
                if not opts.no_EQ_pat:
                    for pat in single_EQ[BQ][mtype]:
                        single_mut, single_nomut = single_EQ[BQ][mtype][pat]
                        double_mut, double_nomut =  double_EQ[BQ][mtype][pat]    
                        beta_single = (alpha*(1.0-avg_single_EQ[BQ][mtype]))/avg_single_EQ[BQ][mtype]
                        beta_double = (alpha*(1.0-avg_double_EQ[BQ][mtype]))/avg_double_EQ[BQ][mtype]
                        s_EQ = (single_mut + alpha) / (single_mut + single_nomut + alpha +beta_single)
                        d_EQ = (double_mut + alpha) / (double_mut + double_nomut + alpha +beta_double)
                        rel_EQ = s_EQ / d_EQ
                        if opts.min1EQ:
                            rel_EQ = max(1.0, rel_EQ)
                        if not opts.output_file_EQ is None:
                            kmp_alpha, kmp_beta = kmer_patterns[BQ][mtype][pat]
                            print(BQ, mtype, pat, kmp_alpha, kmp_beta, kmp_alpha/(kmp_alpha+kmp_beta), single_mut, single_nomut, double_mut, double_nomut, s_EQ, d_EQ, subtract, rel_EQ, file = opts.output_file_EQ)
                        for kmer in matches(pat):
                            kmer_papas[BQ_pair][mtype][kmer] = min(kmer_papas[BQ][mtype][kmer] / rel_EQ, 0.25)
                
                else:
                    for kmer in kmer_papas[BQ][mtype]:
                        kmer_papas[BQ_pair][mtype][kmer] = min(kmer_papas[BQ][mtype][kmer] / rel_EQ, 0.25)
                
            elif opts.estimated == "double":
                kmer_papas[BQ][mtype] = {}
                for kmer in kmer_papas[BQ_pair][mtype]:
                    kmer_papas[BQ][mtype][kmer] = min(kmer_papas[BQ_pair][mtype][kmer] * rel_EQ, 0.25)
    
    if not opts.output_file_EQ is None:
        opts.output_file_EQ.close() 

    # Estimate error rates for pairs of different BQs:
    for i in range(len(BQs)):
        for j in range(i+1, len(BQs)):
            BQ1 = BQs[i]
            BQ2 = BQs[j]
            BQ_pair = f'({BQ1},{BQ2})'
            BQ1_pair = f'({BQ1},{BQ1})'
            BQ2_pair = f'({BQ2},{BQ2})'
            kmer_papas[BQ_pair] = {}
            for mtype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T'):
                kmer_papas[BQ_pair][mtype] = {}
                for kmer in kmer_papas[BQ1_pair][mtype]:
                    if opts.mean_type == 'geometric':
                        kmer_papas[BQ_pair][mtype][kmer] = \
                            sqrt(kmer_papas[BQ1_pair][mtype][kmer]*kmer_papas[BQ2_pair][mtype][kmer])
                    elif opts.mean_type == 'harmonic':
                        kmer_papas[BQ_pair][mtype][kmer] = \
                            1.0/(((1.0/kmer_papas[BQ1_pair][mtype][kmer]) + (1.0/kmer_papas[BQ2_pair][mtype][kmer]))/2)
                    elif opts.mean_type == 'arithmetic':
                        kmer_papas[BQ_pair][mtype][kmer] = \
                            (kmer_papas[BQ1_pair][mtype][kmer] + kmer_papas[BQ2_pair][mtype][kmer])/2

    if not opts.output_file_kmerpapa is None:
        for BQ in kmer_papas:
            for mtype in kmer_papas[BQ]:
                for kmer in kmer_papas[BQ][mtype]:
                    print(BQ, mtype, kmer, kmer_papas[BQ][mtype][kmer], file=opts.output_file_kmerpapa)                   
        opts.output_file_kmerpapa.close()
    
    return kmer_papas


def phred_scale_kmerpapas(kmer_papas):
    for bqual in kmer_papas:
        for mtype in kmer_papas[bqual]:
            for context in kmer_papas[bqual][mtype]:
                alpha, beta = kmer_papas[bqual][mtype][context]
                kmer_papas[bqual][mtype][context] = -10*log10(alpha/(alpha+beta))


def run_validation(opts, kmer_papas):
    if opts.verbosity > 0:
        eprint("Printing validation data")
    validator = \
        MutationValidator(
            opts.bam_file, 
            opts.filter_bam_file, 
            opts.validation_bam_file, 
            opts.twobit_file, 
            kmer_papas)
    if opts.chrom is None:
        validator.call_all_chroms()
    else:
        validator.call_mutations(opts.chrom, opts.start, opts.end)

def run_list_validation(opts, kmer_papas):
    if opts.verbosity > 0:
        eprint("Printing validation data")
    validator = \
        ListMutationValidator(
            opts.bam_file, 
            opts.filter_bam_file, 
            opts.validation_list_file, 
            opts.twobit_file, 
            kmer_papas)
    if opts.chrom is None:
        validator.call_all_chroms()
    else:
        validator.call_mutations(opts.chrom, opts.start, opts.end)



def run_adjust(opts, kmer_papas):
    if opts.verbosity > 0:
        eprint("Adjusting base qualities")
    phred_scale_kmerpapas(kmer_papas)
    adjuster = \
        BaseAdjuster(
            opts.bam_file,
            opts.twobit_file, 
            kmer_papas,
            opts.outbam,
            opts.output_adjustments,
            opts.cutoff
        )
    
    n_corrections, n_corrected_reads, n_uncorrected_reads, n_filtered = \
        adjuster.call_mutations(opts.chrom, opts.start, opts.end)

    if opts.verbosity > 0:
        eprint(f'corrected {n_corrections} base qualities in {n_corrected_reads} reads')
        eprint(f'{n_uncorrected_reads} reads were written with no corrections')
        eprint(f'{n_filtered} reads were filtered')


def run_call(opts, kmer_papas):
    if opts.verbosity > 0:
        eprint("Calling somatic variants")
    if opts.prior_N < 0:
        opts.no_update = True
    caller = \
        SomaticMutationCaller(
            opts.bam_file,
            opts.filter_bam_file,
            opts.twobit_file, 
            kmer_papas,
            opts.outfile,
            opts.method,
            opts.cutoff,
            opts.prior_N,
            opts.no_update,
            opts.min_MQ,
            opts.min_BQ,
            opts.filter_max_count,
            opts.pop_vcf,
            opts.min_enddist,
            opts.max_mismatch,
            opts.max_NM_diff,
            bed_file = opts.exclude_bed,
        )
    if opts.chrom is None:
        n_calls = caller.call_all_chroms()
    else:
        n_calls = caller.call_mutations(opts.chrom, opts.start, opts.end)

    if opts.verbosity > 0:
        if opts.cutoff is None:
            eprint(f'Found {n_calls} possible variants.')
        else:
            eprint(f'Found {n_calls} variants with a quality above {opts.cutoff}.')
    

def get_phred(tup):
    alpha, beta = tup
    return -10*log10(alpha /(alpha+beta))

def run_test_kmerpapas(opts, kmer_papas, good_kmers, bad_kmers):
    for bqual in kmer_papas:
        for mtype in kmer_papas[bqual]:
            if mtype[0] == mtype[-1]:
                continue
            eprint(f'Handling base_qual: {bqual}, mutation type: {mtype}')
            for pat in kmer_papas[bqual][mtype]:
                if opts.correction_type == "bad_vs_good":
                    n_bad = sum(bad_kmers[bqual][mtype][x] for x in matches(pat))
                    if opts.same_good:
                        n_good = sum(good_kmers[37][mtype][x] for x in matches(pat))
                    else:
                        n_good = sum(good_kmers[bqual][mtype][x] for x in matches(pat))
                elif opts.correction_type == "bad_vs_no":
                    ref = mtype[0]
                    alt = mtype[-1]
                    notype = f'{ref}->{ref}'
                    other_type1, other_type2 = [f'{ref}->{other}' for other in 'ACGT' if other != alt and other != ref]
                    n_bad = sum(bad_kmers[bqual][mtype][x] for x in matches(pat))
                    if opts.same_good:
                        n_good = sum(good_kmers[37][notype][x] for x in matches(pat))
                    else:
                        n_good = sum(good_kmers[bqual][notype][x] + bad_kmers[bqual][other_type1][x] + bad_kmers[bqual][other_type2][x] for x in matches(pat))

                if n_bad > 0:
                    phred = -10*log10(n_bad / (n_bad + n_good))
                else:
                    phred = 0
                print(bqual, mtype, pat, get_phred(kmer_papas[bqual][mtype][pat]), n_bad, n_good, phred)


def run_calls_filter(opts):
    if opts.verbosity > 0:
        eprint(f'Filtering PASS calls based on {opts.lower_q*100}% and {opts.upper_q*100}% coverage quantiles.')

    filter = \
        BBQFilter(
            opts.vcf_file, 
            opts.outfile, 
            opts.lower_q,
            opts.upper_q
            )
    n_filtered = filter.filter_BBQ()
    
    if opts.verbosity > 0:
        eprint(f"Removed {n_filtered} PASS calls.")


def main(args = None):
    """
    Run the main program.

    This function is executed when you type `BetterBaseQuals` or `python -m betterbasequals`.

    Arguments:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    opts = parser.parse_args(args=args)

    if "region" in opts:
        parse_opts_region(opts)

    if opts.command is None:
        parser.print_help(sys.stderr)
        return 1

    if not opts.command in ['train_only', 'validate_only', 'list_validate_only', 'call_only', 'adjust_only', 'test_kmerpapa', 'filter_calls']:
        event_kmers = run_get_good_and_bad_w_filter(opts)
    elif opts.command in ['train_only', 'test_kmerpapa']:
        event_kmers = read_kmers(opts)

    if opts.command == 'count':
        return 0

    if not opts.command in ['validate_only', 'list_validate_only', 'call_only', 'adjust_only', 'test_kmerpapa', 'filter_calls']:
        kmer_papas = run_get_kmerpapas(opts, event_kmers)
    elif opts.command in "test_kmerpapa":
        kmer_papas = read_kmer_papas_for_test(opts)
    elif opts.command in "filter_calls":
        pass
    else:
        eprint("Reading kmer pattern partitions")
        kmer_papas = read_kmer_papas(opts)

    if opts.command in ['train_only', 'train']:
        return 0

    if opts.command in ['validate', 'validate_only']:
        run_validation(opts, kmer_papas)
    elif opts.command in ['list_validate', 'list_validate_only']:
        run_list_validation(opts, kmer_papas)
    elif opts.command in ['call', 'call_only']:
        run_call(opts, kmer_papas)
    #elif opts.command in ['adjust', 'adjust_only']:
    #    run_adjust(opts, kmer_papas)
    #elif opts.command == 'test_kmerpapa':
    #    run_test_kmerpapas(opts, kmer_papas, event_kmers)
    elif opts.command == 'filter_calls':
        run_calls_filter(opts)
    else:
        eprint("Unknown command: {opts.command}")
        return 1

    return 0





from betterbasequals.utils import eprint
from kmerpapa.algorithms import greedy_penalty_plus_pseudo
from kmerpapa.algorithms import bottum_up_array_penalty_plus_pseudo_CV
from kmerpapa.algorithms import bottum_up_array_w_numba
from kmerpapa.pattern_utils import get_M_U, generality
from math import log10
from importlib import reload
import argparse

def get_kmerpapa(super_pattern, contextD, name, opts):
    n_bad = 0
    n_good = 0
    for x,y in contextD.values():
        n_bad += x
        n_good += y
    my=n_bad/(n_good+n_bad)
    if opts.verbosity > 0:
        eprint(f' .  n_bad = {n_bad}, n_good = {n_good}, mean_error_rate = {my}')
    if n_bad<10:
        assert n_good > 10, f'too few counts'
        if opts.verbosity > 0:
            eprint(' .  Skipping kmerpapa because of low number of bad kmers')
        return {super_pattern: (0.01+n_bad, 1+n_good)}
        #return {super_pattern: -10*log10((0.5+n_bad)/(0.5+n_good +n_bad))}
    if n_good<10:
        assert n_bad > 10, f'too few counts'
        if opts.verbosity > 0:
            eprint(' .  skipping kmerpapa because of low number of good kmers')
        return {super_pattern: (0.01+n_bad, 1+n_good)}
        #return {super_pattern: -10*log10((n_bad)/(0.5+n_good +n_bad))}
    if opts.kmerpapa_method == 'greedy':
        CV = greedy_penalty_plus_pseudo.BaysianOptimizationCV(super_pattern, contextD, opts.nfolds, opts.iterations, opts.seed, min_pseudo=0.5, min_penalty=1.0)
        best_alpha, best_penalty, test_score = CV.get_best_a_c()
        #eprint(test_score)
        best_beta = (best_alpha*(1.0-my))/my
        best_score, M, U, names = greedy_penalty_plus_pseudo.greedy_partition(super_pattern, contextD, best_alpha, best_beta, best_penalty, opts)
    elif opts.kmerpapa_method == 'optimal':
        args = argparse.Namespace()
        args.nfolds = opts.nfolds
        args.iterations = opts.iterations
        args.verbosity = opts.verbosity
        if opts.CVfile_prefix is None:
            args.CVfile = None
        else:
            cvfile = open(f'{opts.CVfile_prefix}_{name}.txt', 'w')
            args.CVfile = cvfile
        args.seed = opts.seed
        pseudo_counts = opts.pseudo_counts
        penalty_values = opts.penalty_values
        #Right now it is necessary to reload kmerpapa as it uses global variables with numba.
        #If if remove the global variables in kmerpapa I can skip these lines.
        reload(bottum_up_array_penalty_plus_pseudo_CV)
        reload(bottum_up_array_w_numba)
        if len(pseudo_counts) == 1 and len(penalty_values) == 1:
            best_alpha = pseudo_counts[0]
            best_penalty = penalty_values[0]
        else:
            best_alpha, best_penalty, test_score = bottum_up_array_penalty_plus_pseudo_CV.pattern_partition_bottom_up(super_pattern, contextD, pseudo_counts, args, n_bad, n_good, penalty_values)
        best_beta = (best_alpha*(1.0-my))/my
        best_score, M, U, names = bottum_up_array_w_numba.pattern_partition_bottom_up(super_pattern, contextD, best_alpha, best_beta, best_penalty, opts, n_bad, n_good, index_mut=0)
        if not args.CVfile is None:
            cvfile.close()
    if opts.verbosity > 0:
        eprint(f' .  best_penalty = {best_penalty}, best_alpha={best_alpha}, n_patterns={len(names)}')
    counts = []
    for pat in names:
        counts.append(get_M_U(pat, contextD))
    D = {}
    x = 0
    for i in range(len(names)):
        pat = names[i]
        M, U = counts[i]
        #p = (M + best_alpha)/(M + U + best_alpha + best_beta)
        #eprint(pat, p)
        #D[pat] = -10*log10(p)
        D[pat] = (M + best_alpha, U + best_beta) # alpha and beta in posterior beta distribution
        x += generality(pat)
    assert x == generality(super_pattern)
    return D



import csv
import sys
#from itertools import product, count
#from operator import add
#from functools import reduce, partial
from collections import defaultdict, Counter
from math import log10
import pysam
import os
import gzip

change_mtypes = ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T')
mtypes = ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T', 'A->A', 'C->C')


complement = str.maketrans("ATCGN", "TAGCN")

def reverse_complement(s):
    return s.translate(complement)[::-1]

def p2phred(p):
    return -10*log10(p)

def phred2p(Q):
    return 10**(-Q/10)

def empirical_bayes(a, b, x, N):
    return (a+x)/(a+b+N)
    #return p2phred(a+x) - p2phred(a+b+N)

ostrand = {'A':'T', 'C':'G', 'G':'C', 'T':'A'}
def mut_type(from_base, to_base, kmer):
    center = len(kmer)//2
    new_kmer = kmer[:center] + from_base + kmer[center+1:]
    if from_base in ['G','T']:
        return f'{ostrand[from_base]}->{ostrand[to_base]}', reverse_complement(new_kmer)
    else:
        return f'{from_base}->{to_base}', new_kmer

def SW_type(base):
    if base in ['G','T']:
        return ostrand[base]
    else:
        return base

def first(the_iterable):
    for x in the_iterable:
        return x

def get_average_coverage(bamfile):
    """Quickly calculates the average coverage of a WGS bamfile using the bam index.
    OBS: Only works for bam not cram!

    Args:
        bamfile (pysam.AlignmentFile): bamfile

    Returns:
        int: average coverage over all cromosomes where at least one reads map.
    """
    read_len = 0
    i = 0
    for read in bamfile.fetch():
        read_len = max(read_len, read.query_length)
        i += 1
        if i > 10:
            break
    n_mapped = 0
    genome_len = 0
    for idx_stats in bamfile.get_index_statistics():
        print(idx_stats)
        if idx_stats.mapped > 0:
            n_mapped += idx_stats.mapped
            genome_len += bamfile.get_reference_length(idx_stats.contig)

    return (n_mapped * read_len)/genome_len

code = {'A':['A'],
        'C':['C'],
        'G':['G'],
        'T':['T'],
        'R':['A', 'G'],
        'Y':['C', 'T'],
        'S':['G', 'C'],
        'W':['A', 'T'],
        'K':['G', 'T'],
        'M':['A', 'C'],
        'B':['C', 'G', 'T'],
        'D':['A', 'G', 'T'],
        'H':['A', 'C', 'T'],
        'V':['A', 'C', 'G'],
        'N':['A', 'C', 'G', 'T']}

def matches(pattern):
    if len(pattern) == 0:
        yield ''
    else:
        for y in matches(pattern[1:]):
            for x in code[pattern[0]]:
                yield x+y


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

class Read:
    def __init__(self, pileup_read):
        self.pos = pileup_read.query_position

        # set attributes
        self.start = pileup_read.alignment.reference_start
        self.end = pileup_read.alignment.reference_end

        if pileup_read.is_del or pileup_read.is_refskip:
            self.allel = '-'
            self.base_qual = 0
        else:
            self.allel = pileup_read.alignment.query_sequence[self.pos]
            self.base_qual = pileup_read.alignment.query_qualities[self.pos]
            self.enddist = min(self.pos, len(pileup_read.alignment.query_sequence)-self.pos)

        self.is_reverse = pileup_read.alignment.is_reverse
        self.query_name = pileup_read.alignment.query_name
        self.length = abs(pileup_read.alignment.template_length)
        self.isR1 = pileup_read.alignment.is_read1
        self.mapq = pileup_read.alignment.mapping_quality
        #self.NH = pileup_read.alignment.get_tag("NH")
        # Process cigar stats
        cigar_stats = pileup_read.alignment.get_cigar_stats()[0]
        self.has_indel = sum(cigar_stats[1:4]) != 0
        self.has_clip = sum(cigar_stats[4:6]) != 0
        #self.NM = cigar_stats[8]
        self.NM = cigar_stats[10]
        #print(pileup_read.alignment.get_cigar_stats())
    
    def is_good(self, min_enddist=6, max_mismatch = 2, min_mapq = 50):
        return (self.NM <= max_mismatch and
                not self.has_indel and
                not self.has_clip and
                self.enddist >= min_enddist and
                self.mapq >= min_mapq and
                self.allel in ['A', 'C', 'G', 'T'])


class ReadPair:
    def __init__(self, read1, read2):
        self.length = read1.length
        self.allel = read1.allel
        self.start = read2.start if read1.is_reverse else read1.start
        self.min_base_qual = min(read1.base_qual, read2.base_qual)
        self.max_NM = max(read1.NM, read2.NM)
        self.has_indel = 1 if (read1.has_indel or read2.has_indel) else 0
        self.has_clip = 1 if (read1.has_clip or read2.has_clip) else 0



def zip_pileups(*pileups):
    """Iterate through a number of pilup objects and get the columns that are in all.
    Assumes chromosomes are sorted lexicographically in all files
    Yields:
        list : list of pileup columns objects with identical reference pos.
    """
    pileupcolumns = [None] * len(pileups)
    max_chrom = "0"
    while True:
        try:
            for i in range(len(pileups)):
                pileupcolumns[i]  = next(pileups[i])
            while True:
                chrom = max(x.reference_name for x in pileupcolumns)
                assert chrom >= max_chrom, f"chromosomes should be sorted lexicographically, prev:{max_chrom}, current:{chrom}"
                max_chrom = chrom
                max_pos = max(x.reference_pos for x in pileupcolumns)
                for i in range(len(pileups)):
                    while pileupcolumns[i].reference_name < max_chrom:
                        pileupcolumns[i] = next(pileups[i])
                for i in range(len(pileups)):
                    while pileupcolumns[i].reference_name == max_chrom and pileupcolumns[i].reference_pos < max_pos:
                        pileupcolumns[i] = next(pileups[i])
                if all(x.reference_pos == pileupcolumns[0].reference_pos and x.reference_name == pileupcolumns[0].reference_name for x in pileupcolumns):
                    yield pileupcolumns
                    break
        except StopIteration:
            break

def zip_pileups_single_chrom(*pileups):
    """Iterate throug a number of pilup objects and get the columns that are in all.
    Assumes that all pilups are on same chromosome. 
    Yields:
        list : list of pileup columns objects with identical reference pos.
    """
    pileupcolumns = [None] * len(pileups)
    while True:
        try:
            for i in range(len(pileups)):
                pileupcolumns[i]  = next(pileups[i])
            while True:
                max_pos = max(x.reference_pos for x in pileupcolumns)
                for i in range(len(pileups)):
                    while pileupcolumns[i].reference_pos < max_pos:
                        pileupcolumns[i] = next(pileups[i])
                if all(x.reference_pos == pileupcolumns[0].reference_pos for x in pileupcolumns):
                    yield pileupcolumns
                    break
        except StopIteration:
            break


def read_variant_set(variant_file):
    f = open(variant_file)
    S = set()
    for line in f:
        chrom, pos, allele = line.split()
        S.add((chrom, int(pos), allele))
    return S
    #chrom2set = {}
    #for line in f:
    #    chrom, pos, allele = line.split()
    #    if chrom not in chrom2set:
    #        chrom2set[chrom] = set()
    #        chrom2set[chrom].add((int(pos), allele))
    #return chrom2set


def open_bam_w_index(bam_file):
    """Check if bam index exists, if not create an index file.
    Then open and return pysam.AlignmentFile.

    Args:
        bam_file (str): file name

    Returns:
        _type_: _description_
    """
    bai_filename1 = f"{bam_file}.bai"
    bai_filename2 = bam_file[:-1] + 'i'
    if not os.path.exists(bai_filename1) and not os.path.exists(bai_filename2):
        print(f"No index file found ({bai_filename1}), generating...")
        pysam.index(bam_file)
    return pysam.AlignmentFile(bam_file, "rb")

# def read_kmers(opts):
#     if opts.verbosity > 0:
#         eprint("Reading good and bad kmers")
#     good_kmers = {}
#     for line in opts.input_file_good:
#         bqual, mtype, kmer, count = line.split()
#         bqual = int(bqual)
#         if bqual not in good_kmers:
#             good_kmers[bqual] = {}
#             for muttype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T', 'C->C', 'A->A'):
#                 good_kmers[bqual][muttype] = defaultdict(int)
#         good_kmers[bqual][mtype][kmer] = int(count)

#     bad_kmers = {}
#     for line in opts.input_file_bad:
#         bqual, mtype, kmer, count = line.split()
#         if count == "0":
#             continue
#         bqual = int(bqual)
#         if bqual not in bad_kmers:
#             bad_kmers[bqual] = {}
#             for muttype in ('A->C', 'A->G', 'A->T', 'C->A', 'C->G', 'C->T'):
#                 bad_kmers[bqual][muttype] = defaultdict(int)
#         bad_kmers[bqual][mtype][kmer] = int(count)
#     return good_kmers, bad_kmers

def read_kmers(opts):
    event_kmers = defaultdict(int)
    if opts.verbosity > 0:
        eprint("Reading kmer counts")
    for line in opts.input_file_kmers:
        event_type, bqual, mtype, kmer, count = line.split()
        event_kmers[(event_type, bqual, mtype, kmer)] = int(count)
    return event_kmers


# def read_kmer_papas(opts):
#     kmer_papas = {}
#     for line in opts.input_file_kmerpapa:
#         bqual, mtype, pattern, correction_factor = line.split()
#         bqual=int(bqual)
#         if bqual not in kmer_papas:
#             kmer_papas[bqual] = {}
#         if mtype not in kmer_papas[bqual]:
#             kmer_papas[bqual][mtype] = {}
#         for context in matches(pattern):
#             kmer_papas[bqual][mtype][context] = float(correction_factor)
#     return kmer_papas

# def read_kmer_papas_for_test(opts):
#     kmer_papas = {}
#     for line in opts.input_file_kmerpapa:
#         bqual, mtype, pattern, correction_factor = line.split()
#         bqual=int(bqual)
#         if bqual not in kmer_papas:
#             kmer_papas[bqual] = {}
#         if mtype not in kmer_papas[bqual]:
#             kmer_papas[bqual][mtype] = {}
#         kmer_papas[bqual][mtype][pattern] = float(correction_factor)
#     return kmer_papas

# def read_kmer_papas(opts):
#     kmer_papas = {}
#     for line in opts.input_file_kmerpapa:
#         bqual, mtype, pattern, alpha, beta = line.split()
#         alpha = float(alpha)
#         beta = float(beta)
#         bqual=int(bqual)
#         if bqual not in kmer_papas:
#             kmer_papas[bqual] = {}
#         if mtype not in kmer_papas[bqual]:
#             kmer_papas[bqual][mtype] = {}
#         for context in matches(pattern):
#             kmer_papas[bqual][mtype][context] = (alpha, beta)
#     return kmer_papas

def read_kmer_papas(opts):
    kmer_papas = {}
    for line in opts.input_file_kmerpapa:
        bqual, mtype, kmer, p_error = line.split()
        p_error = float(p_error)
        if bqual not in kmer_papas:
            kmer_papas[bqual] = {}
        if mtype not in kmer_papas[bqual]:
            kmer_papas[bqual][mtype] = {}
        kmer_papas[bqual][mtype][kmer] = p_error
    return kmer_papas

def read_kmer_papas_for_test(opts):
    kmer_papas = {}
    for line in opts.input_file_kmerpapa:
        bqual, mtype, pattern, alpha, beta = line.split()
        alpha = float(alpha)
        beta = float(beta)
        bqual=int(bqual)
        if bqual not in kmer_papas:
            kmer_papas[bqual] = {}
        if mtype not in kmer_papas[bqual]:
            kmer_papas[bqual][mtype] = {}
        kmer_papas[bqual][mtype][pattern] = p2phred(alpha/(alpha+beta))
    return kmer_papas


def print_kmer_papas(opts, kmer_papas):
    if not opts.output_file_kmerpapa is None:
        for bqual in kmer_papas:
            for mtype in kmer_papas[bqual]:
                for pat in kmer_papas[bqual][mtype]:
                    Q = kmer_papas[bqual][mtype][pat]
                    print(bqual, mtype, pat, Q, file=opts.output_file_kmerpapa)
        opts.output_file_kmerpapa.close()


def print_kmer_counts(opts, event_kmers):
    if not opts.output_file_kmers is None:
        for tup, count in event_kmers.items():
            print(" ".join(str(x) for x in tup), count, file = opts.output_file_kmers)
        opts.output_file_kmers.close()


# def print_good_and_bad(opts, good_kmers, bad_kmers):
#     if not opts.output_file_good is None:
#         for bqual in good_kmers:
#             for mtype in mtypes:
#                 super_pattern = 'N'*opts.radius + mtype[0] + 'N'*opts.radius
#                 for kmer in matches(super_pattern):
#                     print(bqual, mtype, kmer, good_kmers[bqual][mtype][kmer] , file = opts.output_file_good)
#         opts.output_file_good.close()
#     if not opts.output_file_bad is None:
#         for bqual in bad_kmers:
#             for mtype in mtypes:
#                 super_pattern = 'N'*opts.radius + mtype[0] + 'N'*opts.radius
#                 for kmer in matches(super_pattern):
#                     print(bqual, mtype, kmer, bad_kmers[bqual][mtype][kmer], file = opts.output_file_bad)
#         opts.output_file_bad.close()

# def print_good_and_bad(opts, good_kmers, bad_kmers, single_kmers):
#     if not opts.output_file_good is None:
#         for tup, count in good_kmers.items():
#             bqual, mtype, kmer = tup
#             print(bqual, mtype, kmer, count , file = opts.output_file_good)
#         opts.output_file_good.close()
#     if not opts.output_file_bad is None:
#         for tup, count in bad_kmers.items():
#             bqual, mtype, kmer = tup
#             print(bqual, mtype, kmer, count , file = opts.output_file_bad)
#         opts.output_file_bad.close()
#     if not opts.output_file_single is None:
#         for tup, count in single_kmers.items():
#             bqual, mtype, kmer = tup
#             print(bqual, mtype, kmer, count , file = opts.output_file_single)
#         opts.output_file_single.close()


def parse_opts_region(opts):
    if opts.region is None:
        opts.chrom = None
        opts.start = None
        opts.end = None
    elif ":" in opts.region:
        opts.chrom, end_points = opts.region.split(':')
        opts.start, opts.end = end_points.split('-')
        opts.start = int(opts.start)
        opts.end = int(opts.end)
    else:
        opts.chrom = opts.region
        opts.start = None
        opts.end = None


class VcfAfReader:
    """A simple VCF-file reader that only parses and returns AF field
    for each position
    """ 

    def __init__(self, vcfname, ignore_indels=True, field_name="AF"):
        # TODO: check if name ends in gz and then open using gzip
        if vcfname.endswith('.gz') or vcfname.endswith('.bgz') or vcfname.endswith('.zip'):
            self.f = gzip.open(vcfname, 'rt')
        else:
            self.f = open(vcfname)
        
        # Read lines until we are past comment lines
        vcf_line = self.f.readline()
        while vcf_line.startswith("#"):
            vcf_line = self.f.readline()
        
        vcfL = vcf_line.split()
        self.vcf_chrom = vcfL[0]
        self.vcf_pos = int(vcfL[1])
        self.vcf_ref = vcfL[3]
        self.vcf_alt = vcfL[4]
        self.vcf_info = vcfL[7]
        self.ignore_indels = ignore_indels
        self.field_name = field_name
        self.last_pos = -1
        self.last_chrom = "0"

    def parse_line(self, vcf_line):
        vcfL = vcf_line.split()
        if len(vcfL) == 0:
            self.vcf_chrom = 'z' # should sort after all posible query chrom names
            return

        if vcfL[0] == self.vcf_chrom:
            assert int(vcfL[1]) >= self.vcf_pos, "ERROR: vcffile positions not sorted!"
        else:
            assert vcfL[0] >= self.vcf_chrom, "ERROR: vcffile chromosomes not sorted!"

        self.vcf_chrom = vcfL[0]
        self.vcf_pos = int(vcfL[1])
        self.vcf_ref = vcfL[3]
        self.vcf_alt = vcfL[4]
        self.vcf_info = vcfL[7]

    def parse_AF(self):
        fn = self.field_name + '='
        L = self.vcf_info.split(";")
        for x in L:
            if x.startswith(fn):
                return float(x[len(fn):])

    def query(self, chrom, pos, ref):
        #Check that current query is after the last query
        assert chrom > self.last_chrom or (chrom == self.last_chrom and pos > self.last_pos), "ERROR: queries should be in sorted order! Please sort bam file."
        self.last_chrom = chrom
        self.last_pos = pos

        while self.vcf_chrom < chrom or (self.vcf_chrom == chrom and self.vcf_pos < pos):
            self.parse_line(self.f.readline())
       
        if (chrom == self.vcf_chrom) and ( self.vcf_pos == pos):
            af_dict = defaultdict(float)
            while((chrom == self.vcf_chrom) and ( self.vcf_pos == pos)):

                #Check that reference matches
                assert self.vcf_ref[0] == ref, f'''
                    ERROR: ref allele in vcffile ({self.vcf_ref[0]}) doesn't match ref({ref}).
                    line: {self.vcf_chrom} {self.vcf_pos} {self.vcf_ref} {self.vcf_alt} {self.vcf_info}.
                    Check genome build!
                    '''

                #ignore indels if ignore_indels option is set:
                if not self.ignore_indels or len(self.vcf_ref)==1 and len(self.vcf_alt)==1:
                    af_dict[self.vcf_alt] = self.parse_AF()

                self.parse_line(self.f.readline())

            return af_dict
        else:
            return None


def print_base_probs(base_probs):
    for A in base_probs:
        print(f"P_{A}->X", f"P_R->X", "count", sep='\t')
        D = Counter()
        for x in base_probs[A]:
            D[x] += 1
        for x in D:
            p_A2X, p_R2X, MQ = x
            print(phred2p(p_A2X),phred2p(p_R2X), D[x], sep = '\t')

class SortedBed:
    def __init__(self, bedname):
        self.name = bedname
        self.f = open(bedname)
        self.bed_chrom = 'chr0'
        self.bed_start = -1
        self.bed_end = -1

    def query(self, chrom, pos):
        # TODO: check if the input is sorted
        while self.bed_chrom < chrom or (self.bed_chrom == chrom and self.bed_end <= pos):
            bedL = self.f.readline().split()
            if len(bedL) == 0:
                break
            if (bedL[0] == self.bed_chrom and
                ((int(bedL[1])+1) < self.bed_start or
                 (int(bedL[2])+1) < self.bed_end)):
                eprint("bed positions not sorted:", self.name)
                eprint(' '.join(bedL))
                eprint( self.bed_chrom, self.bed_start, self.bed_end)
                sys.exit()
            if bedL[0] < self.bed_chrom:
                eprint("bed chromosomes not sorted:", self.name)
                eprint(' '.join(bedL))
                eprint(self.bed_chrom, self.bed_start, self.bed_end)
                sys.exit()
            self.bed_chrom = bedL[0]
            self.bed_start = int(bedL[1])+1
            self.bed_end = int(bedL[2])+1
        return (chrom == self.bed_chrom) and ( self.bed_start <= pos) and (pos < self.bed_end)



from collections import Counter, defaultdict
from betterbasequals.utils import Read, ReadPair, reverse_complement, phred2p, p2phred, mut_type
import math

def get_mut_type(ref, papa_ref, alt):
    if ref != papa_ref:
        mtype = papa_ref + '->' + reverse_complement(alt)
    else:
        mtype = papa_ref + '->' + alt
    return mtype


def get_pileup_count(pileupcolumn, ref, min_base_qual, blood=False):
    n_ref = 0
    n_alt = Counter()
    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None

            #or read.NH != 1
        ):
            continue

        if (read.base_qual < min_base_qual):
            continue

        if (not blood) and (read.has_clip == 1 or
            read.has_indel == 1 or
            read.NM >3):
            continue
            
        if read.allel == ref:
            n_ref += 1
        else:
            n_alt[read.allel] += 1
    return n_ref, n_alt



def get_pileup_count_double(pileupcolumn, ref, min_base_qual):
    n_ref = 0
    n_alt = Counter()
    reads_mem = dict()
    has_incompatible = {x:False for x in 'ACGT'}
    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue

        # Look for read partner
        if read.query_name in reads_mem:
            # found partner process read pair
            mem_read = reads_mem[read.query_name]

            # Test if readpair is okay
            if (
                read.is_reverse == mem_read.is_reverse
                or read.length != mem_read.length
                or min(read.base_qual, mem_read.base_qual) < min_base_qual
            ):
                continue

            read_pair = ReadPair(read, mem_read)

            #check read_pair filters
            if (read_pair.has_clip == 1 or
                read_pair.has_indel == 1 or
                read_pair.max_NM >3):
                continue
            
            if read.allel != mem_read.allel:
                has_incompatible[read.allel] = True
                has_incompatible[mem_read.allel] = True
                continue
                
            # Count alleles
            if read.allel == ref:
                n_ref += 1
            else:
                n_alt[read.allel] += 1
                
        else:
            # Unseen read partner store read
            reads_mem[read.query_name] = read
    return n_ref, n_alt, has_incompatible


def get_alleles_w_corrected_quals(pileupcolumn, ref, papa_ref, kmer, correction_factor):
    """
    Returns a dictionary that maps from alleles to a list of tuples of the form: (adjusted_BQ, old_BQ, type)
    Where type is:
    1) Match in overlap
    2) Mismatch in overlap
    3) Base with no overlap.
    """

    reads_mem = {}
    base_quals = {'A':[], 'C':[], 'G':[], 'T':[]}

    #tmp: counting ref... can be removed later
    n_ref = 0
    n_mismatch = Counter()
    n_double = 0

    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue
        #TODO: should consider what the right solution is if there is deletion at overlap

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue

        # Look for read partner
        if read.query_name in reads_mem:
            # found partner process read pair
            mem_read = reads_mem.pop(read.query_name)

            n_double += 1
            # We ignore alleles where pair doesn't match (could add else clause to handle)
            ## Have now done so to test --- Can maybe delete later
            if read.allel == mem_read.allel:
                mut_type = get_mut_type(ref, papa_ref, read.allel)

                #Are ignoring ref alleles pt... should adjust later
                if read.allel == ref:
                    n_ref +=1
                    continue
                adjusted_base_qual1 = correction_factor[read.base_qual][mut_type][kmer]
                adjusted_base_qual2 = correction_factor[mem_read.base_qual][mut_type][kmer]
                adjusted_base_qual = max(adjusted_base_qual1, adjusted_base_qual2) + 3
                unadjusted_base_qual = max(read.base_qual, mem_read.base_qual)
                base_quals[read.allel].append((adjusted_base_qual, unadjusted_base_qual, 1))
            else:
                
                #Are ignoring ref alleles pt... should adjust later
                if read.allel != ref:
                    n_mismatch[read.allel] += 1
                    base_quals[read.allel].append((0, read.base_qual, 2))
                else:
                    n_ref += 1
                #Are ignoring ref alleles pt... should adjust later
                if mem_read.allel != ref:
                    n_mismatch[mem_read.allel] += 1
                    base_quals[mem_read.allel].append((0, mem_read.base_qual, 2))
                else:
                    n_ref += 1

        else:            
            reads_mem[read.query_name] = read

    # Handle reads without partner (ie. no overlap)
    for read in reads_mem.values():
        if read.allel != ref:
            mut_type = get_mut_type(ref, papa_ref, read.allel)
            adjusted_base_qual = correction_factor[read.base_qual][mut_type][kmer]
            base_quals[read.allel].append((adjusted_base_qual, read.base_qual, 3))
        else:
            n_ref += 1
    return base_quals, n_ref, n_mismatch, n_double


def get_alleles_w_probabities_seperate(pileupcolumn, ref, ref_kmer, correction_factor, improve = 1, ignore_ref = False, double_adjustment="max_plus3"):
    """
    Returns a dictionary that maps from allele A to a list of tuples with probability of 
    observing Alt allele A given then read and probability of observing ref allele R given 
    the read. I.e.: base_probs[A] = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]
    The probabilities are given on phred scale.
    We only considder reads where X_read_i == A or X_read_i == R.
    """

    reads_mem = {}
    base_probs = {'A':[], 'C':[], 'G':[], 'T':[]}
    seen_alt = set()
    n_mismatch = Counter()
    n_double = Counter()
    n_alt = Counter()

    #tmp: counting ref... can be removed later
    n_ref = 0
    R = ref
    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue
        #TODO: should consider what the right solution is if there is deletion at overlap

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue


        # Look for read partner
        if read.query_name in reads_mem:
            # found partner process read pair
            mem_read = reads_mem.pop(read.query_name)

            # We do not trust mismathces in overlaps so we only handle matches
            if read.allel == mem_read.allel:
                X = read.allel
            
                if X == R:
                    if ignore_ref:
                        continue
                    alts = [A for A in ['A','C','G','T'] if A!=R]
                else:
                    alts = [X]
                    seen_alt.add(X)

                for A in alts:
                    n_double[A] += 1
                    
                    muttype_from_A, kmer_from_A = mut_type(A, X, ref_kmer)
                    a1, b1 = correction_factor[read.base_qual][muttype_from_A][kmer_from_A]
                    a2, b2 = correction_factor[mem_read.base_qual][muttype_from_A][kmer_from_A]
                    
                    if b1 is None:
                        adjusted_base_qual_from_A = ((a1+a2)/(2*improve), None)
                    else:
                        adjusted_base_qual_from_A = (a1+a2, improve*(b1+b2))
                    
                    muttype_from_R, kmer_from_R = mut_type(R, X, ref_kmer)
                    a1, b1 = correction_factor[read.base_qual][muttype_from_R][kmer_from_R]
                    a2, b2 = correction_factor[mem_read.base_qual][muttype_from_R][kmer_from_R]
                    
                    if b1 is None:
                        adjusted_base_qual_from_R = ((a1+a2)/(2*improve), None)
                    else:
                        #print("double", a1, a2, b1, b2, muttype_from_R, kmer_from_R)
                        adjusted_base_qual_from_R = (a1+a2, improve*(b1+b2))
                    BQ = max(read.base_qual, mem_read.base_qual)
                    if BQ > 35:
                        n_alt[X] += 1
                    base_probs[A].append((adjusted_base_qual_from_A, adjusted_base_qual_from_R, BQ))

                    # adjusted_base_qual_from_A1 = correction_factor[read.base_qual][muttype_from_A][kmer_from_A]
                    # adjusted_base_qual_from_A2 = correction_factor[mem_read.base_qual][muttype_from_A][kmer_from_A]
                    
                    # if double_adjustment == "mult":
                    #     adjusted_base_qual_from_A = adjusted_base_qual_from_A1 + adjusted_base_qual_from_A2
                    # elif double_adjustment == "max_plus3":
                    #     adjusted_base_qual_from_A = max(adjusted_base_qual_from_A1, adjusted_base_qual_from_A2) + 3

                    # muttype_from_R, kmer_from_R = mut_type(R, X, ref_kmer)
                    # adjusted_base_qual_from_R1 = correction_factor[read.base_qual][muttype_from_R][kmer_from_R]
                    # adjusted_base_qual_from_R2 = correction_factor[mem_read.base_qual][muttype_from_R][kmer_from_R]
                    
                    # if double_adjustment == "mult":
                    #     adjusted_base_qual_from_R = adjusted_base_qual_from_R1 + adjusted_base_qual_from_R2
                    # elif double_adjustment == "max_plus3":
                    #     adjusted_base_qual_from_R = max(adjusted_base_qual_from_R1, adjusted_base_qual_from_R2) + 3
                        
                    # base_probs[A].append((adjusted_base_qual_from_A, adjusted_base_qual_from_R))
            else: # Mismatch
                if read.allel != ref:
                    n_mismatch[read.allel] += 1
                    n_double[read.allel] += 1
                if mem_read.allel != ref:
                    n_mismatch[mem_read.allel] += 1
                    n_double[mem_read.allel] += 1
        else:            
            reads_mem[read.query_name] = read

    # Handle reads without partner (ie. no overlap)
    for read in reads_mem.values():
        X = read.allel
            
        if X == R:
            alts = [A for A in ['A','C','G','T'] if A!=R]
        else:
            alts = [X]
            seen_alt.add(X)
            if read.base_qual > 35:
                n_alt[X] += 1
        
        for A in alts:
            muttype_from_A, kmer_from_A = mut_type(A, X, ref_kmer)
            adjusted_base_qual_from_A = correction_factor[read.base_qual][muttype_from_A][kmer_from_A]
            
            muttype_from_R, kmer_from_R = mut_type(R, X, ref_kmer)
            adjusted_base_qual_from_R = correction_factor[read.base_qual][muttype_from_R][kmer_from_R]
            
            base_probs[A].append((adjusted_base_qual_from_A, adjusted_base_qual_from_R, read.base_qual))
    
    posterior_base_probs = {'A':[], 'C':[], 'G':[], 'T':[]}
    for A in base_probs:
        for (from_A, from_R, BQ) in base_probs[A]:
            a_from_A, b_from_A  = from_A
            a_from_R, b_from_R  = from_R 
            
            ## TODO: Should I do bayesian update on both from_A and from_R
            ## Or only on from_R ?????

            ## should I also save alpha and beta for posterior? 
            ## Or is it enough that I just save mean?

            if b_from_A is None:
                posterior_from_A = a_from_A
            else:
                posterior_from_A = (a_from_A + n_mismatch[A])/(a_from_A + b_from_A + n_double[A])

            if b_from_R is None:
                posterior_from_R = a_from_R
            else:
                posterior_from_R = (a_from_R + n_mismatch[A])/(a_from_R + b_from_R + n_double[A])

            posterior_base_probs[A].append((posterior_from_A, posterior_from_R))

    return base_probs, posterior_base_probs, seen_alt, n_mismatch, n_double, n_alt


# def get_alleles_w_probabities_update_ver2(pileupcolumn, ref, ref_kmer, correction_factor, prior_N, no_update, double_adjustment=0.5):
#     """
#     Returns a dictionary that maps from allele A to a list of tuples with probability of 
#     observing Alt allele A given then read and probability of observing ref allele R given 
#     the read. I.e.: base_probs[A] = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]
#     The probabilities are given on phred scale.
#     We only considder reads where X_read_i == A or X_read_i == R.
#     """

#     reads_mem = {}
#     seen_alt = set()
#     n_mismatch = Counter()
#     n_mismatch_BQ = {'A':Counter(), 'C':Counter(), 'G':Counter(), 'T':Counter()}
#     n_double = Counter()
#     n_double_BQ = {'A':Counter(), 'C':Counter(), 'G':Counter(), 'T':Counter()}
#     n_alt = Counter()
#     events = {'A':[], 'C':[], 'G':[], 'T':[]}

#     R = ref
#     for pileup_read in pileupcolumn.pileups:
#         # test for deletion at pileup
#         if pileup_read.is_del or pileup_read.is_refskip:
#             continue
#         #TODO: should consider what the right solution is if there is deletion at overlap

#         # fetch read information
#         read = Read(pileup_read)

#         if not read.is_good():
#             continue

#         # test if read is okay
#         if (
#             read.allel not in "ATGC"
#             or read.start is None
#             or read.end is None
#             #or read.NH != 1
#         ):
#             continue


#         # Look for read partner
#         if read.query_name in reads_mem:
#             # found partner process read pair
#             mem_read = reads_mem.pop(read.query_name)

#             # We do not trust mismathces in overlaps so we only add to events list in case of match
#             if read.allel == mem_read.allel:
#                 X = read.allel
            
#                 if X == R:
#                     alts = [A for A in ['A','C','G','T'] if A!=R]
#                 else:
#                     alts = [X]
#                     seen_alt.add(X)

#                 for A in alts:
#                     #if not no_update:    
#                     n_double[A] += 1

#                     read_MQ = (read.mapq + mem_read.mapq)/2
#                     #if overlap_type == "double":
#                     read_BQ = max(read.base_qual, mem_read.base_qual)
#                     events[A].append(("double", X, read_BQ, read_MQ))

#             else: # Mismatch
#                 #if not no_update:
#                 # TODO: Would it make sense to also count ref_mismatches so that we could
#                 # do bayesian update of A->R error rates and not only R->A error rates?
#                 if read.allel != ref and mem_read.allel == ref:
#                     n_mismatch[read.allel] += 1
#                     n_mismatch_BQ[read.allel][read.base_qual] += 1
#                     for A in ['A','C','G','T']:
#                         if A == ref:
#                             continue
#                         n_double[A] += 1
#                         n_double_BQ[A][read.base_qual] += 1
#                 if mem_read.allel != ref and read.allel == ref:
#                     n_mismatch[mem_read.allel] += 1
#                     n_mismatch_BQ[mem_read.allel][mem_read.base_qual] += 1
#                     for A in ['A','C','G','T']:
#                         if A == ref:
#                             continue
#                         n_double[A] += 1
#                         n_double_BQ[A][mem_read.base_qual] += 1
#         else:            
#             reads_mem[read.query_name] = read

#     # Handle reads without partner (ie. no overlap)
#     for read in reads_mem.values():
#         X = read.allel
            
#         if X == R:
#             alts = [A for A in ['A','C','G','T'] if A!=R]
#         else:
#             alts = [X]
#             seen_alt.add(X)
        
#         for A in alts:
#             events[A].append(("single", X, read.base_qual, read.mapq))
        
#     new_correction_factor = defaultdict(dict)

#     #I only need to calculate probabilities of changing bases from one of the seen alleles.
#     # I have to considder change to all bases to calculate stay types (X->X) correctly.
#     relevant_bases = [ref] + list(seen_alt)

#     for BQ in correction_factor:
#         new_correction_factor[BQ] = defaultdict(dict)

#         for from_base in relevant_bases:
#             p_rest = 1.0
#             stay_type, stay_kmer = mut_type(from_base, from_base, ref_kmer)
#             for to_base in ['A', 'C', 'G', 'T']:
#                 if to_base == from_base:
#                     continue
#                 change_type, change_kmer = mut_type(from_base, to_base, ref_kmer)
#                 alpha, beta = correction_factor[BQ][change_type][change_kmer]
#                 p_prior = alpha/(alpha+beta)
#                 new_p_prior = 0
#                 for other_BQ in correction_factor:
#                     other_alpha, other_beta = correction_factor[BQ][change_type][change_kmer]
#                     other_p_prior = other_alpha/(other_alpha, other_beta)
#                     new_p_prior += BQ_freq[BQ] * 
#                 a = p_prior * prior_N
#                 b = prior_N - a
#                 if no_update:
#                     p_posterior = a/(a + b)
#                 else: 
#                     p_posterior = (a + n_mismatch[to_base])/(a + b + n_double[to_base])
#                 p_rest -= p_posterior
                
#                 new_correction_factor[BQ][change_type][change_kmer] = p2phred(p_posterior)
           
#             new_correction_factor[BQ][stay_type][stay_kmer] = p2phred(p_rest)
 
#     # calculate error rates for double reads:
#     for BQ1, BQ2 in double_combinations:
#         new_correction_factor[(BQ1,BQ2)] = defaultdict(dict)
#         for from_base in relevant_bases:
#             p_rest = 1.0
#             stay_type, stay_kmer = mut_type(from_base, from_base, ref_kmer)
#             for to_base in ['A', 'C', 'G', 'T']:
#                 if to_base == from_base:
#                     continue
#                 change_type, change_kmer = mut_type(from_base, to_base, ref_kmer)
#                 alpha1, beta1 = correction_factor[BQ1][change_type][change_kmer]
#                 alpha2, beta2 = correction_factor[BQ2][change_type][change_kmer]
                
#                 p_prior_1 = alpha1/(alpha1+beta1)
#                 p_prior_2 = alpha2/(alpha2+beta2)

#                 p_prior_double = math.exp(math.log(p_prior_1)/double_adjustment) * p_prior_2
#                 a = p_prior_double * prior_N
#                 b = prior_N - a
#                 if no_update or from_base != ref:
#                     p_posterior = a/(a + b)
#                 else:
#                     p_posterior = (a + n_mismatch)/(a + b + n_double)
#                     #p_posterior = (a + n_mismatch[to_base])/(a + b + n_double[to_base])
#                 #print(ref, BQ, from_base, to_base, p_posterior, n_mismatch[to_base], n_double[to_base])
#                 p_rest -= p_posterior
#                 new_correction_factor[(BQ1, BQ2)][change_type][change_kmer] = p2phred(p_posterior)
#             new_correction_factor[(BQ1, BQ2)][stay_type][change_kmer] = p2phred(p_rest)



#     posterior_base_probs = {'A':[], 'C':[], 'G':[], 'T':[]}
#     BQs = {'A':[], 'C':[], 'G':[], 'T':[]}
#     for A in seen_alt:
#         for overlap_type, X, read_BQ, read_MQ in events[A]:
#             #muttype_from_A, kmer_from_A = mut_type(A, X, ref_kmer)
#             muttype_from_R, kmer_from_R = mut_type(R, X, ref_kmer)
            
#             #posterior_from_A = new_correction_factor[read_BQ][overlap_type][muttype_from_A][kmer_from_A]
#             posterior_from_R = new_correction_factor[read_BQ][overlap_type][muttype_from_R][kmer_from_R]
#             if A==X:
#                 posterior_base_probs[A].append((posterior_from_R, read_MQ, read_BQ))
#             #if A==X:
#             #    BQs[A].append((read_BQ, posterior_from_R))
                    
#     return posterior_base_probs, n_mismatch, n_double, n_mismatch_BQ, n_double_BQ




def get_validation_probabities(pileupcolumn, ref, ref_kmer, correction_factor, improve = 1, ignore_ref = False, double_adjustment="max_plus3"):
    """
    
    """

    reads_mem = {}
    base_probs = {'A':[], 'C':[], 'G':[], 'T':[]}
    seen_alt = set()
    n_mismatch = Counter()
    n_double = Counter()
    n_alt = Counter()

    #tmp: counting ref... can be removed later
    n_ref = 0
    R = ref
    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue
        #TODO: should consider what the right solution is if there is deletion at overlap

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue


        # Look for read partner
        if read.query_name in reads_mem:
            # found partner process read pair
            mem_read = reads_mem.pop(read.query_name)

            # We do not trust mismathces in overlaps so we only handle matches
            if read.allel == mem_read.allel:
                A = read.allel
                
                if A == R:
                    for X in [x for x in 'ACGT' if x != R]:
                        n_double[X] += 1
                    continue
                else:
                    n_double[A] += 1
                seen_alt.add(A)

                muttype_from_R, kmer_from_R = mut_type(R, A, ref_kmer)
                

                BQ = max(read.base_qual, mem_read.base_qual)

                alpha, beta = correction_factor[BQ][muttype_from_R][kmer_from_R]                
                
                if BQ > 35:
                    n_alt[A] += 1

                base_probs[A].append((alpha, beta, BQ, muttype_from_R, 1))

            else: # Mismatch
                if read.allel != ref:
                    n_mismatch[read.allel] += 1
                    n_double[read.allel] += 1
                if mem_read.allel != ref:
                    n_mismatch[mem_read.allel] += 1
                    n_double[mem_read.allel] += 1
        else:            
            reads_mem[read.query_name] = read

    # Handle reads without partner (ie. no overlap)
    for read in reads_mem.values():
        A = read.allel
                
        if A == R:
            continue

        seen_alt.add(A)
        if read.base_qual > 35:
            n_alt[A] += 1

        muttype_from_R, kmer_from_R = mut_type(R, A, ref_kmer)
        alpha, beta = correction_factor[read.base_qual][muttype_from_R][kmer_from_R]
        BQ = read.base_qual
        base_probs[A].append((alpha, beta, BQ, muttype_from_R, 2))
    
    #final_base_probs = {'A':[], 'C':[], 'G':[], 'T':[]}
    #for A in base_probs:
    #    for alpha, beta, BQ, muttype, atype in base_probs[A]:        
    #        final_base_probs[A].append((alpha, beta, BQ, muttype, atype, n_mismatch[A], n_))

    return base_probs, seen_alt, n_mismatch, n_double, n_alt





def get_adjustments(pileupcolumn, ref, papa_ref, kmer, correction_factor, change_dict):
    reads_mem = {}

    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue
        #TODO: should consider what the right solution is if there is deletion at overlap

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue

        # Look for read partner
        if read.query_name in reads_mem:
            # found partner process read pair
            mem_read = reads_mem.pop(read.query_name)

            # overlap matches
            if read.allel == mem_read.allel:
                mut_type = get_mut_type(ref, papa_ref, read.allel)

                #Are ignoring ref alleles pt... should adjust later
                if read.allel == ref:
                    continue

                adjusted_base_qual1 = correction_factor[read.base_qual][mut_type][kmer]
                adjusted_base_qual2 = correction_factor[mem_read.base_qual][mut_type][kmer]

                #adjusted_base_qual1 = read.base_qual + correction_factor[mut_type][kmer]
                #adjusted_base_qual2 = mem_read.base_qual + correction_factor[mut_type][kmer]
                #adjusted_base_qual = adjusted_base_qual1 + adjusted_base_qual2

                change_dict[(read.query_name, read.isR1)].append((read.pos, read.base_qual, read.allel, int(adjusted_base_qual1), 1))
                change_dict[(mem_read.query_name, mem_read.isR1)].append((mem_read.pos, mem_read.base_qual, mem_read.allel, int(adjusted_base_qual2), 1))

            else: # overlap mismatch

                #Are ignoring ref alleles pt... could adjust later
                if read.allel != ref:
                    change_dict[(read.query_name, read.isR1)].append((read.pos, read.base_qual, read.allel, 0, 2))
                #Are ignoring ref alleles pt... could adjust later
                if mem_read.allel != ref:
                    change_dict[(mem_read.query_name, mem_read.isR1)].append((mem_read.pos, mem_read.base_qual, mem_read.allel, 0, 2))
        else:            
            reads_mem[read.query_name] = read

    # Handle reads without partner (ie. no overlap)
    for read in reads_mem.values():
        if read.allel != ref:
            mut_type = get_mut_type(ref, papa_ref, read.allel)
            adjusted_base_qual = correction_factor[read.base_qual][mut_type][kmer]
            change_dict[(read.query_name, read.isR1)].append((read.pos, read.base_qual, read.allel, int(adjusted_base_qual), 3))



def get_alleles_w_quals(pileupcolumn):
    base_quals = {'A':[], 'C':[], 'G':[], 'T':[]}
    for pileup_read in pileupcolumn.pileups:
        # test for deletion at pileup
        if pileup_read.is_del or pileup_read.is_refskip:
            continue

        # fetch read information
        read = Read(pileup_read)

        if not read.is_good():
            continue

        # test if read is okay
        if (
            read.allel not in "ATGC"
            or read.start is None
            or read.end is None
            #or read.NH != 1
        ):
            continue
        base_quals[read.allel].append(read.base_qual)

    return base_quals




from betterbasequals.utils import eprint
from operator import itemgetter
import numpy as np


class BBQFilter:
    def __init__(
        self,
        vcf_file,
        outfile,
        cutoff_lower, 
        cutoff_upper
    ):

        self.vcf = vcf_file
        self.cutoff_lower = cutoff_lower
        self.cutoff_upper = cutoff_upper
        self.outfile = outfile


    def filter_BBQ(self):
        cov_list = []
        cov_total_list = []
        calls = open(self.vcf, 'r')
        n_pass = 0

        # create a list of PASS variant coverages
        for line in calls:
            if line.split("\t")[6] == "PASS":
                n_pass += 1
                cov = itemgetter(1, 4)(line.split("\t")[7].split(";"))
                cov_list.append(int(cov[0].split("=")[1]))
                cov_total_list.append(int(cov[1].split("=")[1]))
        calls.close()
        eprint(f'Number of PASS calls before filtering: {n_pass}')
        
        cov_arr = np.array(cov_list)
        cov_total_arr = np.array(cov_total_list)

        # filtered coverage quantiles of PASS variants
        cov_q_l = np.quantile(cov_arr, self.cutoff_lower)
        cov_q_u = np.quantile(cov_arr, self.cutoff_upper)

        # total coverage quantiles of PASS variants
        cov_total_q_l = np.quantile(cov_total_arr, self.cutoff_lower)
        cov_total_q_u = np.quantile(cov_total_arr, self.cutoff_upper)

        # filter PASS variants based on coverage quantiles 
        pass_count = 0
        n_filtered = 0
        n_pass = 0
        with open(self.vcf, 'r') as calls:
            for line in calls:
                if line.split("\t")[6] == "PASS":
                    cov = cov_list[pass_count]
                    cov_total = cov_total_list[pass_count]
                    pass_count += 1
                    if cov_q_l <= cov <= cov_q_u:
                        if cov_total_q_l <= cov_total <= cov_total_q_u:
                            n_pass += 1
                            self.outfile.write(line)
                        else: 
                            n_filtered += 1
                    else: 
                        n_filtered += 1
                else: 
                    self.outfile.write(line)

        eprint(f'Number of PASS calls after filtering: {n_pass}')
        
        return n_filtered
                


    



import py2bit
from scipy.optimize import minimize_scalar
from betterbasequals.utils import eprint, zip_pileups_single_chrom, open_bam_w_index, phred2p, p2phred, VcfAfReader, mut_type, Read, change_mtypes, SW_type, SortedBed, first
from collections import Counter, defaultdict
import scipy.stats
import math

def GQ_sum(alt_BQs, ref_BQs):
    return sum(alt_BQs)

def get_LR(base_probs):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    def p_data_given_mut(alpha):
        return - sum(math.log(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) for p_a2x, p_r2x, _ in base_probs)
        
        # If probabilities in base_probs are not phred scaled. It becomes this:
        #return sum(p2phred(alpha * p_a2x + (1-alpha)*p_r2x) for p_a2x, p_r2x in base_probs)
        
    res = minimize_scalar(p_data_given_mut, bounds=(0, 1), method='bounded')

    alpha = res.x
    LR = - res.fun + p_data_given_mut(0)
    return LR, alpha


def get_maxLR_with_MQ(base_probs):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    LL_with_alt = sum(math.log((1-phred2p(p_map_error))*max(phred2p(p_a2x),phred2p(p_r2x)) + phred2p(p_map_error)*0.5) for p_a2x, p_r2x, p_map_error in base_probs if p_r2x > p_a2x)
    LL_no_alt = sum(math.log((1-phred2p(p_map_error))*phred2p(p_r2x) + phred2p(p_map_error)*0.5) for p_a2x, p_r2x, p_map_error in base_probs if p_r2x > p_a2x)

    LR = LL_with_alt - LL_no_alt
    return LR


def get_LR_with_MQ(base_probs):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    def p_data_given_mut(alpha):
        #return sum(p2phred(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) for p_a2x, p_r2x in base_probs)
        return - sum(math.log((1-phred2p(p_map_error))*(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) + phred2p(p_map_error)*0.5) for p_a2x, p_r2x, p_map_error in base_probs)

    res = minimize_scalar(p_data_given_mut, bounds=(0, 1), method='bounded')
    alpha = res.x
    LR = - res.fun + p_data_given_mut(0)
    return LR, alpha


def get_BF(base_probs):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    def p_data_given_mut(alpha):
        return phred2p(sum(p2phred(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) for p_a2x, p_r2x, _ in base_probs))
        
    # Integrate out alpha 
    p_no_alpha, error = scipy.integrate.quad(p_data_given_mut, 0, 1)

    try:
        log_BF = math.log(p_no_alpha) - math.log(p_data_given_mut(0))
    except:
        log_BF = -1000

    return log_BF

def get_BF_with_MQ(base_probs):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    def p_data_given_mut(alpha):
        return phred2p(sum(p2phred((1-phred2p(p_map_error))*(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) + phred2p(p_map_error)*0.5) for p_a2x, p_r2x, p_map_error in base_probs))

    # Integrate out alpha
    p_no_alpha, error = scipy.integrate.quad(p_data_given_mut, 0, 1)

    try:
        log_BF = math.log(p_no_alpha) - math.log(p_data_given_mut(0))
    except:
        log_BF = -1000

    return log_BF

def get_BF_with_MQ_and_Prior(base_probs, a=1, b=2):
    #base_probs = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

    def p_data_given_mut(alpha):
        return scipy.stats.beta.pdf(alpha, a, b) * phred2p(sum(p2phred((1-phred2p(p_map_error))*(alpha * phred2p(p_a2x) + (1-alpha)*phred2p(p_r2x)) + phred2p(p_map_error)*0.5) for p_a2x, p_r2x, p_map_error in base_probs))

    # Integrate out alpha
    p_no_alpha, error = scipy.integrate.quad(p_data_given_mut, 0, 1)

    try:
        log_BF =  math.log(p_no_alpha) - math.log(p_data_given_mut(0))
    except:
        log_BF = - 1000

    return log_BF


class SomaticMutationCaller:
    def __init__(
        self,
        bam_file,
        filter_bam_file,
        twobit_file,
        kmer_papa,
        outfile,
        method,
        cutoff,
        prior_N,
        no_update,
        mapq = 50,
        min_base_qual = 1,
        min_filter_count = 2,
        pop_vcf = None,
        min_enddist = 6,
        max_mismatch = 2,
        max_NM_diff = 1,
        filter_mapq = 20,
        min_base_qual_filter = 20, 
        min_depth = 1, 
        max_depth = 1000000,
        #radius = 3, 
        prefix = "",
        bed_file = None,
    ):

        # Open files for reading
        self.bam_file = open_bam_w_index(bam_file)
        if filter_bam_file is None:
            self.filter_bam_file = None
        else:
            self.filter_bam_file = open_bam_w_index(filter_bam_file)

        self.tb = py2bit.open(twobit_file)

        # assumes that the kmer papa has been turned into phred scaled probabilities
        self.mut_probs = kmer_papa

        self.outfile = outfile
        self.method = method

        self.cutoff = cutoff
        self.prior_N = prior_N
        self.no_update = no_update

        if not pop_vcf is None:
            self.pop_vcf = VcfAfReader(pop_vcf)
        else:
            self.pop_vcf = None

        self.min_enddist = min_enddist
        self.max_mismatch = max_mismatch
        self.max_NM_diff = max_NM_diff
        self.mapq = mapq
        self.min_base_qual = min_base_qual
        self.filter_mapq = filter_mapq
        self.min_base_qual_filter = min_base_qual_filter
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.radius = len(first(first(first(kmer_papa.values()).values()).keys()))//2
        self.prefix = prefix
        self.filter_variants = False

        #TODO: filter_depth variable should be set based on average coverage in filter file.
        self.min_filter_depth = 10
        self.max_filter_depth = 1000000
        self.min_filter_count = min_filter_count

        if not bed_file is None:
            self.bed = SortedBed(bed_file)
        else:
            self.bed = None
                

    def __del__(self):
        self.tb.close()

    def call_all_chroms(self):
        n_calls = 0
        for idx_stats in self.bam_file.get_index_statistics():
            if idx_stats.mapped > 0:
                n_calls += self.call_mutations(idx_stats.contig, None, None)
        return n_calls

    def call_mutations(self, chrom, start, stop):
        pileup = self.bam_file.pileup(
            contig = chrom,
            start = start,
            stop = stop,
            truncate = True,
            max_depth = 1000000,
            min_mapping_quality = 0, #self.mapq,
            ignore_overlaps = False,
            flag_require = 0,  # No requirements
            flag_filter = 3840,
            min_base_quality = self.min_base_qual,
        )
        n_calls = 0
        if not self.filter_bam_file is None:
            filter_pileup = self.filter_bam_file.pileup(
                contig = chrom,
                start = start,
                stop = stop,
                truncate = True,
                min_mapping_quality = self.filter_mapq,
                ignore_overlaps = False,
                flag_require = 2,  # proper paired
                flag_filter = 3840,
                min_base_quality = self.min_base_qual_filter,
            )
            for pileupcolumn, filter_pc in zip_pileups_single_chrom(pileup, filter_pileup):
                n_alleles = Counter()
                for pread in filter_pc.pileups:
                    pos = pread.query_position
                    if not pos is None:
                        n_alleles[pread.alignment.query_sequence[pos]] += 1
                N_filter = sum(n_alleles.values())
                if N_filter < self.min_filter_depth or N_filter > self.max_filter_depth:
                    continue
                #filter_alleles = [x for x,y in n_alleles.items() if y >= self.min_filter_count]
                n_calls += self.handle_pileup(pileupcolumn, n_alleles)
        else:
            for pileupcolumn in pileup:
                n_calls += self.handle_pileup(pileupcolumn, None)
        return n_calls


    def handle_pileup(self, pileupcolumn, n_alleles):
        ref_pos = pileupcolumn.reference_pos
        chrom = pileupcolumn.reference_name
        if ref_pos%100000 == 0:
            eprint(f"{chrom}:{ref_pos}")

        if not self.bed is None and self.bed.query(chrom, ref_pos):
            return 0

        n_calls = 0

        if ref_pos-self.radius < 0:
            return 0
        
        kmer = self.tb.sequence(self.prefix + chrom, ref_pos- self.radius, ref_pos + self.radius + 1)
        if 'N' in kmer:
            return 0
        if len(kmer) != 2*self.radius + 1:
            return 0
        ref = kmer[self.radius]
        if ref not in "ATGC":
            return 0

        if self.method == 'poisson':
            raise NotImplementedError
            #make new pileup handler
        elif self.method == 'sum':
            raise NotImplementedError
            #sum = sum(from_R for from_A,from_R in base_probs[A] if from_A < from_R)
        elif self.method in ['LR','LR_with_MQ', 'BF', 'BF_with_MQ', 'BF_with_MQ_and_Prior', 'maxLR_with_MQ']:
            base_probs, BQs, n_mismatch, n_double, n_pos, n_neg, n_filtered, n_nonfiltered = \
                self.get_alleles_w_probabities_update(pileupcolumn, ref, kmer)
            #base_probs[A] = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]

            #no_filter_base_probs, no_filter_BQs, no_filter_n_mismatch, no_filter_n_double, no_filter_n_pos, no_filter_n_neg = \
            #    get_alleles_w_probabities_update(pileupcolumn, ref, kmer, self.mut_probs, self.prior_N, self.no_update, self.double_adjustment, filter_reads=False)

            if not self.pop_vcf is None:
                pop_af = self.pop_vcf.query(chrom, ref_pos+1, ref)
            else:
                pop_af = None

            for A in base_probs:
                F_list = []
                
                if not n_alleles is None and n_alleles[A] > self.min_filter_count:
                    continue
                #if A in filter_alleles:
                #    continue
                altMQs = [MQ for x,y,MQ in base_probs[A] if x<y]
                if len(altMQs) == 0:
                    continue
                altMQs.sort()
                medianMQ = altMQs[len(altMQs)//2]
                
                N_total = sum(n_filtered.values()) + sum(n_nonfiltered.values())

                if medianMQ < 30:
                    if self.filter_variants:
                        continue
                    else:
                        F_list.append("lowMQ")

                refMQs = [MQ for x,y,MQ in base_probs[A] if x>y]
                refMQs.sort()
                
                if len(refMQs)>0:
                    ref_medianMQ = refMQs[len(refMQs)//2]
                else:
                    ref_medianMQ = 0
                    continue

                AF = None
                if self.method == 'LR':
                    QUAL, AF = get_LR(base_probs[A])
                elif self.method == 'LR_with_MQ':
                    QUAL, AF = get_LR_with_MQ(base_probs[A])
                elif self.method == 'maxLR_with_MQ':
                    QUAL = get_maxLR_with_MQ(base_probs[A])
                elif self.method == 'BF':
                    QUAL = get_BF(base_probs[A])
                elif self.method == 'BF_with_MQ':
                    QUAL = get_BF_with_MQ(base_probs[A])
                elif self.method == 'BF_with_MQ_and_Prior':
                    QUAL = get_BF_with_MQ_and_Prior(base_probs[A])

                if self.cutoff is None or QUAL >= self.cutoff:

                    N = len(base_probs[A])
                    N_A = sum(1 for x,y,_ in base_probs[A] if x<y)
                    if AF is None:
                        AF = N_A/N

                    oldBQ_str = '[' + ','.join(x[-1] for x in BQs[A]) + ']'
                    newBQ = '[' +','.join(f'{x[1]:.1f}' for x in BQs[A]) + ']'
                    
                    oldBQ = [x[0] for x in BQs[A]]
                    oldBQ.sort()
                    
                    medianBQ = oldBQ[len(oldBQ)//2]
                    
                    n37 = sum(x[0]==37 for x in BQs[A])
                    a = n_filtered[A]
                    b = n_nonfiltered[A]
                    c = n_filtered[ref]
                    d = n_nonfiltered[ref]
                    if a*d <= b*c:
                        filter_allele_pval = 1.0
                    else:
                        filter_allele_pval = scipy.stats.fisher_exact([[a,b],[c,d]], alternative='greater')[1]
                    filter_allele_table = f'[{a},{b},{c},{d}]'

                    if filter_allele_pval < 0.05:
                        if self.filter_variants:
                            continue
                        else:
                            F_list.append("FilterBias")

                    indel_reads = n_filtered['-']

                    if medianBQ < 20:
                        if self.filter_variants:
                            continue
                        else:
                            F_list.append("lowBQ")

                    if min(n_pos[A], n_neg[A]) == 0 and max(n_pos[A], n_neg[A])>4:
                        if self.filter_variants:
                            continue
                        else:
                            F_list.append("StrandBias")


                    enddist = [x[2] for x in BQs[A]]
                    enddist_str = str(enddist)
                    enddist.sort()
                    median_enddist = enddist[len(enddist)//2]

                    if median_enddist <= 1:
                        if self.filter_variants:
                            continue
                        else:
                            F_list.append("lowEndDist")

                    has_indel = [x[3] for x in BQs[A]]
                    has_clip = [x[4] for x in BQs[A]]
                    NM_alt = [x[5] for x in BQs[A]]
                    NM_ref = [x[5] for x in BQs[ref]]
                    #NM_str = str(NM)
                    NM_alt.sort()
                    NM_ref.sort()
                    median_NM_alt = NM_alt[len(NM_alt)//2]
                    min_NM_alt = min(NM_alt)

                    if len(NM_ref)>0:
                        median_NM_ref = NM_ref[len(NM_ref)//2]
                        min_NM_ref = min(NM_ref)
                        quartile_NM_ref = NM_ref[(len(NM_ref)*3)//4]
                    else:
                        median_NM_ref = 0
                        min_NM_ref = 0
                    
                    #if min_NM_alt - median_NM_ref > self.max_NM_diff:
                    if median_NM_alt - median_NM_ref > self.max_NM_diff:
                        if self.filter_variants:
                            continue
                        else:
                            F_list.append("manyAltMismatches")

                    frac_indel = sum(has_indel)/len(has_indel)
                    frac_clip = sum(has_clip)/len(has_clip)

                    
                    n37_other = sum(x[0]==37 for alt in 'ACGT' for x in BQs[alt] if alt not in [ref,A])
                    #n37_other_nf = sum(x[0]==37 for alt in 'ACGT' for x in no_filter_BQs[alt] if alt not in [ref,A])

                    n_calls += 1
                    if len(F_list) == 0:
                        FILTER = "PASS"
                    else:
                        FILTER = ','.join(F_list)

                    optional_info = ''
                    if not n_alleles is None:
                        optional_info = f';N_A_f={n_alleles[A]}'
                    if not pop_af is None and pop_af[A] > 1e-7:
                        optional_info += f';pop_AF={pop_af[A]}'

                    #print(f'{chrom}\t{ref_pos+1}\t.\t{ref}\t{A}\t{QUAL}\t{FILTER}\tpval={p_val:.3g};LR={LR:.3f};AF={AF:.3g};N={N};N_A={N_A};oldBQ={oldBQ_str};newBQ={newBQ};n_mismatch={n_mismatch[A]};n_overlap={n_double[A]};MQ={int(medianMQ)}', file=self.outfile)
                    #print(f'{chrom}\t{ref_pos+1}\t.\t{ref}\t{A}\t{QUAL}\t{FILTER}\tAF={AF:.3g};N={N};N_A={N_A};N_A_37={n37};oldBQ={oldBQ_str};newBQ={newBQ};n_mismatch={no_filter_n_mismatch[A]};n_overlap={no_filter_n_double[A]};MQ={int(medianMQ)};alt_strand=[{n_pos[A]},{n_neg[A]}];enddist={enddist_str};NM={median_NM};frac_indel={frac_indel:.3g};frac_clip={frac_clip:.3g};kmer={kmer};n_other={n37_other};no_filter_n_other={n37_other_nf}{optional_info}', file=self.outfile)
                    print(f'{chrom}\t{ref_pos+1}\t.\t{ref}\t{A}\t{QUAL:.2f}\t{FILTER}\tAF={AF:.3g};N={N};N_A={N_A};N_A_37={n37};N_total={N_total};oldBQ={oldBQ_str};newBQ={newBQ};n_mismatch={n_mismatch[A]};n_overlap={n_double[A]};MQ={int(medianMQ)};alt_strand=[{n_pos[A]},{n_neg[A]}];enddist={enddist_str};median_alt_NM={median_NM_alt};median_ref_NM={median_NM_ref};min_alt_NM={min_NM_alt};min_ref_NM={min_NM_ref};quartile_NM_ref={quartile_NM_ref};frac_indel={frac_indel:.3g};frac_clip={frac_clip:.3g};kmer={kmer};n_other={n37_other};ref_medianMQ={ref_medianMQ};filter_vs_allele={filter_allele_table};filter_allele_pval={filter_allele_pval:.3g};indel_reads={indel_reads}{optional_info}', file=self.outfile)

        return n_calls
    
    def get_alleles_w_probabities_update(self, pileupcolumn, ref, ref_kmer):
        """
        Returns a dictionary that maps from allele A to a list of tuples with probability of 
        observing Alt allele A given then read and probability of observing ref allele R given 
        the read. I.e.: base_probs[A] = [(P(A -> X_read_i|read_i),P(R -> X_read_i|read_i), ..., ]
        The probabilities are given on phred scale.
        We only considder reads where X_read_i == A or X_read_i == R.
        """

        reads_mem = {}
        seen_alt = set()
        n_mismatch = Counter()
        n_double = Counter()
        n_pos = Counter()
        n_neg = Counter()
        events = {'A':{}, 'C':{}, 'G':{}, 'T':{}}

        n_filtered = Counter()
        n_nonfiltered = Counter()

        observed_BQs = set()
        R = ref
        for pileup_read in pileupcolumn.pileups:
           
            # fetch read information
            read = Read(pileup_read)

            # Look for read partner
            if read.query_name in reads_mem:
                # found partner process read pair
                mem_read = reads_mem.pop(read.query_name)

                # We do not trust mismathces in overlaps so we only add to events list in case of match
                if read.allel == mem_read.allel:
                    X = read.allel

                    if (not read.is_good(self.min_enddist, self.max_mismatch, self.mapq)) and mem_read.is_good(self.min_enddist, self.max_mismatch, self.mapq):
                        #considder mem_read single read.
                        reads_mem[read.query_name] = mem_read
                        continue
                    elif (not mem_read.is_good(self.min_enddist, self.max_mismatch, self.mapq)) and read.is_good(self.min_enddist, self.max_mismatch, self.mapq):
                        #considder read single read.
                        reads_mem[read.query_name] = read
                        continue
                    elif (not read.is_good(self.min_enddist, self.max_mismatch, self.mapq)) and (not mem_read.is_good(self.min_enddist, self.max_mismatch, self.mapq)):
                        #n_filtered += 1
                        n_filtered[X] += 1
                        continue
                    else:
                        n_nonfiltered[X] += 1

                    if X != R:
                        seen_alt.add(X)
                        n_pos[X] += 1
                        n_neg[X] += 1
        
                    #if not no_update:   
                    n_double[X] += 1

                    read_MQ = (read.mapq + mem_read.mapq)/2
                        
                    if read.base_qual > mem_read.base_qual:
                        read_BQ = (read.base_qual, mem_read.base_qual)
                        BQ_pair = f'({read.base_qual},{mem_read.base_qual})'
                    else:
                        read_BQ = (mem_read.base_qual, read.base_qual)
                        BQ_pair = f'({mem_read.base_qual},{read.base_qual})'

                    observed_BQs.add(BQ_pair)
                    enddist = max(read.enddist, mem_read.enddist)
                    has_indel = max(read.has_indel, mem_read.has_indel)
                    has_clip = max(read.has_clip, mem_read.has_clip)
                    NM = max(read.NM, mem_read.NM)
                    fragment_id = (min(read.start, mem_read.start), max(read.end, mem_read.end), X)
                    assert read.start < read.end
                    assert mem_read.start < mem_read.end
                    if fragment_id not in events[X]:
                        events[X][fragment_id] = (read_BQ, read_MQ, enddist, has_indel, has_clip, NM, BQ_pair)
                    else:
                        o_read_BQ = events[X][fragment_id][0]
                        if o_read_BQ < read_BQ:
                            events[X][fragment_id] = (read_BQ, read_MQ, enddist, has_indel, has_clip, NM, BQ_pair)

                else: # Mismatch
                    #if not no_update:
                    # TODO: Would it make sense to also count ref_mismatches so that we could
                    # do bayesian update of A->R error rates and not only R->A error rates?
                    if read.allel != ref and mem_read.allel == ref:
                        n_mismatch[read.allel] += 1
                        for A in ['A','C','G','T']:
                            if A == ref:
                                continue
                            n_double[A] += 1
                    if mem_read.allel != ref and read.allel == ref:
                        n_mismatch[mem_read.allel] += 1
                        for A in ['A','C','G','T']:
                            if A == ref:
                                continue
                            n_double[A] += 1
            else:            
                reads_mem[read.query_name] = read

        # Handle reads without partner (ie. no overlap)
        for read in reads_mem.values():
            X = read.allel
            if not read.is_good(self.min_enddist, self.max_mismatch, self.mapq):
                n_filtered[X] += 1
                continue
            else:    
                n_nonfiltered[X] += 1

            if X != R:
                seen_alt.add(X)
                if read.is_reverse:
                    n_neg[X] += 1
                else:
                    n_pos[X] += 1
            
            observed_BQs.add(str(read.base_qual))
            frag_id = (read.start, X)
            if frag_id not in events[X]:
                events[X][frag_id] = (read.base_qual, read.mapq, read.enddist, read.has_indel, read.has_clip, read.NM, str(read.base_qual))
            else:
                o_read_BQ = events[X][frag_id][0]
                if o_read_BQ < read.base_qual:
                    events[X][frag_id] = (read.base_qual, read.mapq, read.enddist, read.has_indel, read.has_clip, read.NM, str(read.base_qual))                        

        if len(seen_alt) == 0:
            return {}, {}, n_mismatch, n_double, n_pos, n_neg, n_filtered, n_nonfiltered


        new_mut_probs = defaultdict(dict)

        #I only need to calculate probabilities of changing bases from one of the seen alleles.
        # I have to considder change to all bases to calculate stay types (X->X) correctly.
        relevant_bases = [ref] + list(seen_alt)

        # Do positional update of error rates:
        for BQ in observed_BQs:
            new_mut_probs[BQ] = defaultdict(dict)

            for from_base in relevant_bases:
                p_rest = 1.0
                stay_type, stay_kmer = mut_type(from_base, from_base, ref_kmer)
                for to_base in ['A', 'C', 'G', 'T']:
                    if to_base == from_base:
                        continue
                    change_type, change_kmer = mut_type(from_base, to_base, ref_kmer)

                    p_prior = self.mut_probs[BQ][change_type][change_kmer]

                    a = p_prior * self.prior_N
                    b = self.prior_N - a
                    if self.no_update or from_base != ref:
                        p_posterior = a/(a + b)
                    else:
                        p_posterior = (a + n_mismatch[to_base])/(a + b + n_double[to_base])
                    p_rest -= p_posterior

                    assert 0.0 < p_posterior < 1.0, f"posterior error probability outside bounds: {p_posterior}"
                    #print(ref, BQ, from_base, to_base, p_posterior, n_mismatch[to_base], n_double[to_base])
                    new_mut_probs[BQ][change_type][change_kmer] = p2phred(p_posterior)

                assert 0.0 < p_rest < 1.0, f"no change posterior error probability outside bounds: {p_rest}"
                new_mut_probs[BQ][stay_type][stay_kmer] = p2phred(p_rest)    


        posterior_base_probs = {}
        BQs = {'A':[], 'C':[], 'G':[], 'T':[]}        


        for A in seen_alt:
            posterior_base_probs[A] = []
            for read_BQ, read_MQ, enddist, has_indel, has_clip, NM, str_BQ in events[A].values():
                muttype_from_A, kmer_from_A = mut_type(A, A, ref_kmer)
                muttype_from_R, kmer_from_R = mut_type(R, A, ref_kmer)
                posterior_from_A = new_mut_probs[str_BQ][muttype_from_A][kmer_from_A]
                posterior_from_R = new_mut_probs[str_BQ][muttype_from_R][kmer_from_R]

                posterior_base_probs[A].append((posterior_from_A, posterior_from_R, read_MQ))
                #if A==X:
                if type(read_BQ) == tuple:
                    BQ1, BQ2 = read_BQ
                    read_BQ = max(BQ1,BQ2)
                    str_BQ = f'{BQ1}/{BQ2}'                    
                BQs[A].append((read_BQ, posterior_from_R, enddist, has_indel, has_clip, NM, str_BQ))

        for read_BQ, read_MQ, enddist, has_indel, has_clip, NM, str_BQ in events[R].values():
            muttype_from_A, kmer_from_A = mut_type(A, R, ref_kmer)
            muttype_from_R, kmer_from_R = mut_type(R, R, ref_kmer)
            posterior_from_A = new_mut_probs[str_BQ][muttype_from_A][kmer_from_A]
            posterior_from_R = new_mut_probs[str_BQ][muttype_from_R][kmer_from_R]
            for A in seen_alt:
                posterior_base_probs[A].append((posterior_from_A, posterior_from_R, read_MQ))
                #if A==X:
            if type(read_BQ) == tuple:
                BQ1, BQ2 = read_BQ
                read_BQ = max(BQ1,BQ2)
                str_BQ = f'{BQ1}/{BQ2}'
            
            BQs[R].append((read_BQ, posterior_from_R, enddist, has_indel, has_clip, NM, str_BQ))
        

        return posterior_base_probs, BQs, n_mismatch, n_double, n_pos, n_neg, n_filtered, n_nonfiltered





import pysam
import py2bit
import sys
from betterbasequals.utils import reverse_complement, Read, open_bam_w_index
from betterbasequals.pilup_handlers import get_adjustments
from collections import defaultdict

class BaseAdjuster:
    def __init__(
        self,
        bam_file,
        twobit_file,
        kmer_papa,
        output_bam,
        adjustment_file,
        cutoff,
    ):
        # Open files for reading
        self.bam_file = open_bam_w_index(bam_file)

        #Open outputfile for writing
        self.out_file = pysam.AlignmentFile(output_bam, "wb", template=self.bam_file)

        self.tb = py2bit.open(twobit_file)

        #assumes that the kmer papa has been turned into phred scaled correction factor
        self.correction_factor = kmer_papa

        self.adjustment_file = adjustment_file
        self.cutoff = cutoff
                

    def __del__(self):
        self.tb.close()


    def call_all_chroms(self):
        for idx_stats in self.bam_file.get_index_statistics():
            if idx_stats.mapped > 0:
                self.call_mutations(idx_stats.contig, None, None)


    def call_mutations(self, chrom, start, stop, mapq=50, radius=3, prefix=""):
        pileup = self.bam_file.pileup(
            contig=chrom,
            start=start,
            stop=stop,
            truncate=True,
            max_depth = 1000000,
            min_mapping_quality=mapq,
            ignore_overlaps=False,
            flag_require=0,  # proper paired
            flag_filter=3848,
        )
        
        change_dict = defaultdict(list)

        # Fill dict with changes:
        for pileupcolumn in pileup:
            ref_pos = pileupcolumn.reference_pos
            pileup_chrom = pileupcolumn.reference_name            

            try:
                kmer = self.tb.sequence(prefix + pileup_chrom, ref_pos- radius, ref_pos + radius + 1)
            except:
                continue
            
            ref = kmer[radius]
            if 'N' in kmer or len(kmer)!= 2*radius +1:
                continue            

            if kmer[radius] in ['T', 'G']:
                kmer = reverse_complement(kmer)
                papa_ref = reverse_complement(ref)
            else:
                papa_ref = ref

            assert len(ref) == 1
            if ref not in "ATGC":
                continue
            
            get_adjustments(pileupcolumn, ref, papa_ref, kmer, self.correction_factor, change_dict)

        print('n_reads_w_changes:', len(change_dict), file=sys.stderr)
        n_corrected_reads = 0
        n_uncorrected_reads = 0
        n_corrections = 0
        n_filtered = 0

        for read in self.bam_file.fetch(chrom, start, stop):
            if read.is_secondary or read.is_supplementary:
                n_filtered += 1
                continue
            read_id = (read.query_name, read.is_read1)
            if read_id in change_dict:
                n_corrected_reads += 1
                for pos, basequal, allele, adjusted_basequal, atype in change_dict[read_id]:
                    if not self.adjustment_file is None:
                        refpos = read.get_reference_positions(full_length=True)[pos]
                        print(read.reference_name, refpos, allele, basequal, adjusted_basequal, file = self.adjustment_file)
                    n_corrections += 1
                    assert(read.query_sequence[pos] == allele)
                    assert(read.query_qualities[pos] == basequal)
                    if self.cutoff is None or adjusted_basequal >= self.cutoff:
                        read.query_qualities[pos] = adjusted_basequal
                    else:
                        read.query_qualities[pos] = 0
            else:
                n_uncorrected_reads += 1
            self.out_file.write(read)
        return n_corrections, n_corrected_reads, n_uncorrected_reads, n_filtered

